# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# pylint: disable=global-statement
import csv
import logging
import sys
import time
from os import path
from os.path import isfile

import pandas as pd
from bokeh.io import curdoc
from bokeh.layouts import column, layout, WidgetBox
from bokeh.models import ColumnDataSource, Div, Row, ranges, LabelSet, CDSView, IndexFilter
from bokeh.models.widgets import DataTable, TableColumn, RadioGroup, Dropdown, Tabs, Panel, \
    TextInput, Button, Slider
from bokeh.plotting import figure

from nlp_architect.solutions.trend_analysis.trend_analysis import analyze
from nlp_architect import LIBRARY_OUT

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)
dir = str(LIBRARY_OUT / 'trend-analysis-data')
graph_data_path = path.join(dir, 'graph_data.csv')
filter_data_path = path.join(dir, 'filter_phrases.csv')
target_scores_path = path.join(dir, 'target_scores.csv')
ref_scores_path = path.join(dir, 'ref_scores.csv')
max_len = 50
gd = None
top_before = None
top_after = None
hot_trends = None
cold_trends = None
trend_clustering = None
custom_trends = None
topic_clustering_before = None
topic_clustering_after = None
white_list = []
custom_nps = []
ref_data_name = ""
target_data_name = ""
apply_filter = False
dont_regenerate = False
filter_item = -1
all_topics = []
filter_rows = []
table_columns = [
    TableColumn(field="topics", title="Topics")
]


# create ui components

filter_topics_table_source = ColumnDataSource(data={})
view1 = CDSView(source=filter_topics_table_source, filters=[IndexFilter()])
filter_topics_table = DataTable(
    source=filter_topics_table_source, view=view1, columns=table_columns, width=500,
    selectable=True, scroll_to_selection=True, css_classes=['filter_topics_table'])
filter_custom_table_source = ColumnDataSource(data={})
filter_custom_table = DataTable(
    source=filter_custom_table_source, columns=table_columns, width=500, selectable=True,
    css_classes=['filter_custom_table'])
radio_group_area = RadioGroup(
    labels=["All", "Top Topics", "Trends", "Trend Clustering", "Custom Trends", "Topic Clustering",
            "Filter"], active=0, inline=True, width=400)
n_menu = [("5", "5"), ("10", "10"), ("20", "20"), ("30", "30")]
n_clusters_menu = [("20", "20"), ("50", "50"), ("100", "100"), ("300", "300"), ("500", "500")]
scores = [("0", "0"), ("0.1", "0.1"), ("0.25", "0.25"), ("0.5", "0.5"), ("1", "1")]
top_n_dropdown = Dropdown(label="Show Top 10 Phrases", button_type="warning",
                          menu=n_menu, value="10")
top_n_clusters_dropdown = Dropdown(label="Show Top 100 Clusters",
                                   button_type="warning", menu=n_clusters_menu, value="100")
tfidf_slider = Slider(start=0, end=1, value=0, step=.05, title="Informativeness")
cval_slider = Slider(start=0, end=1, value=0, step=.05, title="Completeness")
freq_slider = Slider(start=0, end=1, value=0, step=.05, title="Linguistical")
filter_topics_tab = Panel(child=filter_topics_table, title="Filter Topics")
filter_custom_tab = Panel(child=filter_custom_table, title="Custom Trends")
filter_tabs = Tabs(tabs=[filter_topics_tab, filter_custom_tab])
search_input_box = TextInput(title="Search:", value="", width=300)
search_button = Button(label="Go", button_type="success", width=50, css_classes=['search_button'])
clear_button = Button(label="Clear", button_type="success", width=50, css_classes=['clear_button'])
buttons_layout = column(Div(height=0), Row(search_button, clear_button))
analyse_button = Button(label="re-analyze", button_type="success")
filter_label = Div(
    text="", style={'color': 'red', 'padding-bottom': '0px', 'font-size': '15px'})
label = Div(
    text="", style={'color': 'red', 'padding-bottom': '0px', 'font-size': '15px'})
config_box = WidgetBox(tfidf_slider, cval_slider, freq_slider)
graphs_area = column(children=[])


# define page layout

config_area = layout(
    [
        [radio_group_area],
        [top_n_dropdown],
        [top_n_clusters_dropdown],
        [config_box],
        [analyse_button],


    ])
grid = layout(
    [
        [Div(width=500), Div(text="<H1>AIPG Trend Analysis</H1>")],
        [label],
        [config_area, Div(width=50), graphs_area]
    ]
)


# define functions

def reset_filters():
    """
    Resets the filter tables (e.g. after clearing search)
    """
    logger.info("reset filters")
    global filter_item
    filter_item = -1
    filter_topics_table.view.filters = [IndexFilter()]
    filter_custom_table.view.filters = [IndexFilter()]
    filter_label.text = ""


def refresh_filter_area():
    graphs_area.children = [Row(search_input_box, buttons_layout), filter_label,
                            filter_tabs]


def get_valid_phrases():
    """
    Get a list of all whitelist/valid phrases (which were not filtered-out by the user)
    """
    return [x[0] for x in all_topics if x[1] == '1']


def get_custom_phrases():
    """
    Get a list of all phrases selected by the user for the custom trends graph
    """
    return [x[0] for x in all_topics if x[2] == '1']


def get_valid_indices():
    """
    Get the filter-table indices of all whitelist/valid phrases
    """
    return [i for i, val in enumerate(all_topics) if val[1] == '1']


def is_valid_topic(index):
    """
    Check if a certain phrase should be presented
    """
    return all_topics[index][1] == '1'


def get_custom_indices():
    """
    Get the custom-table indices of all custom selected phrases
    """
    return [i for i, val in enumerate(all_topics) if val[2] == '1']


def search(item):
    """
    Search an item in the filter tables
    """
    row = -1
    try:
        row = filter_rows.index(item)
    except Exception:
        logger.warning("%s was not found", str(item))
    return row


def cut_phrase(phrase):
    """
    Cut phrases which exceeds phrase max length to present
    """
    return phrase if len(phrase) < max_len else phrase[:max_len] + "..."


def read_filter_graph_data():
    global all_topics, filter_rows
    all_topics = []
    filter_rows = []
    with open(filter_data_path) as f:
        reader = csv.reader(f)
        next(reader, None)
        for line in enumerate(reader):
            all_topics.append(line[1])
            filter_rows.append(line[1][0])


def create_topic_clustering_before_graph(top_n):
    if 'Trend Clustering' not in gd['reports'].values:  # no clustering data
        return None
    header = gd['reports'][6]
    scatter_series = gd[(gd['x_ref'] != -1) & (gd['x_ref'].notna())]
    phrases = scatter_series['ref_topic'].values
    x_b = scatter_series['x_ref'].values
    y_b = scatter_series['y_ref'].values
    valid_phrases, valid_x, valid_y = [], [], []
    ctr = 0
    for i in range(len(phrases)):
        if ctr == top_n:
            break
        phrase = phrases[i]
        if phrase in white_list:
            valid_phrases.append(phrase)
            valid_x.append(x_b[i])
            valid_y.append(y_b[i])
            ctr += 1

    source = ColumnDataSource(data=dict(
        x=valid_x,
        y=valid_y,
        topics=valid_phrases
    ))

    p = figure(title=header, plot_width=600, plot_height=400, tooltips='@topics')
    p.circle(x='x', y='y', size=13, color="#335fa5",
             alpha=1, source=source)
    return p


def create_topic_clustering_after_graph(top_n):
    if 'Trend Clustering' not in gd['reports'].values:  # no clustering data
        return None
    header = gd['reports'][7]
    scatter_series = gd[(gd['x_tar'] != -1) & (gd['x_tar'].notna())]
    phrases = scatter_series['tar_topic'].values
    x_a = scatter_series['x_tar'].values
    y_a = scatter_series['y_tar'].values
    valid_phrases, valid_x, valid_y = [], [], []
    ctr = 0
    for i in range(len(phrases)):
        if ctr == top_n:
            break
        phrase = phrases[i]
        if phrase in white_list:
            valid_phrases.append(phrase)
            valid_x.append(x_a[i])
            valid_y.append(y_a[i])
            ctr += 1

    source = ColumnDataSource(data=dict(
        x=valid_x,
        y=valid_y,
        topics=valid_phrases
    ))

    p = figure(title=header, plot_width=600, plot_height=400, tooltips='@topics')
    p.circle(x='x', y='y', size=13, color="#335fa5",
             alpha=1, source=source)

    return p


def regenerate_graphs(top_n_phrases, top_n_clusters):
    logger.info("regenerate graphs")
    global top_before, top_after, hot_trends, cold_trends, trend_clustering, \
        custom_trends, white_list, custom_nps, topic_clustering_before, topic_clustering_after
    white_list = get_valid_phrases()
    custom_nps = get_custom_phrases()
    top_before = create_top_before_graph(int(top_n_phrases))
    top_after = create_top_after_graph(int(top_n_phrases))
    hot_trends = create_trend_graphs(int(top_n_phrases))[0]
    cold_trends = create_trend_graphs(int(top_n_phrases))[1]
    if top_n_clusters != 0:  # no w2v model
        trend_clustering = create_trend_clustering_graph(int(top_n_clusters))
        topic_clustering_before = create_topic_clustering_before_graph(
            int(top_n_clusters))
        topic_clustering_after = create_topic_clustering_after_graph(
            int(top_n_clusters))
    custom_trends = create_custom_trends_graph()


def create_graphs(top_n_phrases, top_n_clusters):
    logger.info("create graphs")
    global topic_clustering_before, topic_clustering_after, gd
    if gd is None:  # initialization
        gd = pd.read_csv(graph_data_path)
        read_filter_graph_data()  # read only once- then update using the filter table
        global white_list, custom_nps
        white_list = get_valid_phrases()
        custom_nps = get_custom_phrases()
        set_weights()
    regenerate_graphs(top_n_phrases, top_n_clusters)


def set_weights():
    weights_series = gd['weights']
    tfidf_slider.value = weights_series[0]
    cval_slider.value = weights_series[1]
    freq_slider.value = weights_series[2]


def re_analyze():
    logger.info("re-analysing scores")
    label.text = "Re-Analysing... Please Wait..."
    global gd
    if isfile(target_scores_path) and isfile(ref_scores_path):
        analyze(target_scores_path, ref_scores_path, tar_header=target_data_name,
                ref_header=ref_data_name,
                re_analysis=True, tfidf_w=float(tfidf_slider.value),
                cval_w=float(cval_slider.value), lm_w=float(freq_slider.value))
        gd = None
        create_graphs(top_n_dropdown.value, int(top_n_clusters_dropdown.value))
        if radio_group_area.active == 0:
            handle_selected_graph(0)
        else:
            radio_group_area.active = 0
        label.text = "Done!"
        time.sleep(2)
        label.text = ""
    else:
        logger.info("scores files are missing")
        label.text = "Error: scores files are missing!"
        time.sleep(2)
    logger.info("re-analysing done")


def create_custom_trends_graph():
    header = gd['reports'][4]
    phrases = gd.values[:, 9]
    imp_change = gd.values[:, 10]
    phrases_h, changes_h = [], []
    phrases_c, changes_c = [], []
    text_h, text_c = [], []
    custom_count = 0
    for i in range(len(phrases)):
        phrase = phrases[i]
        change = float(imp_change[i])
        if phrase in custom_nps:
            if custom_count > 30:
                logger.warning("too many custom phrases (>30). Showing partial list")
                break
            phrase = cut_phrase(phrase)
            if change > 0:
                phrases_h.append(phrase)
                changes_h.append(change)
                text_h.append('+' + str(("%.1f" % change)) + '%')
            else:
                phrases_c.append(phrase)
                changes_c.append(change)
                text_c.append(str(("%.1f" % change)) + '%')
            custom_count += 1

    changes = changes_h + changes_c
    text = text_h + text_c
    trends = phrases_h + phrases_c
    colors = []

    if len(changes_h) > 0:
        for i in range(len(changes_h)):
            colors.append("#1d6d34")
    if len(changes_c) > 0:
        for i in range(len(changes_c)):
            colors.append("#912605")
    if len(changes) < 10:  # pad with 10 blanks
        changes += [0] * (10 - len(changes))
        text += ' ' * (10 - len(text))
        trends += [str((10 - len(trends)) - i) for i in range(0, (10 - len(trends)))]
        colors += ['white'] * (10 - len(colors))

    source = ColumnDataSource(
        dict(y=trends[::-1], x=changes[::-1], colors=colors[::-1], text=text[::-1]))

    plot = figure(title=header, plot_width=600, plot_height=400, tools="save",
                  x_range=ranges.Range1d(start=min(changes) - 10, end=max(changes) + 20),
                  y_range=source.data["y"], tooltips='@y<br>change: @x')
    labels = LabelSet(x=max(changes) + 5, y='y', text='text', level='glyph',
                      x_offset=0, y_offset=-10, source=source,
                      render_mode='canvas')

    plot.hbar(source=source, right='x', y='y', left=0, height=0.5, color='colors')
    plot.add_layout(labels)

    return plot


def create_trend_graphs(top_n):
    header_h = gd['reports'][2]
    header_c = gd['reports'][3]
    phrases = gd.values[:, 9]  # all trends
    imp_change = gd.values[:, 10]
    data_h, data_c = {}, {}
    text_h, text_c = [], []
    hot, cold = 0, 0  # counters for each list
    plot_h = figure(plot_width=600, plot_height=400, title=header_h)
    plot_c = figure(plot_width=600, plot_height=400, title=header_c)
    for i in range(len(phrases)):
        if hot == top_n and cold == top_n:
            break
        phrase = phrases[i]
        change = float(imp_change[i])
        if phrase in white_list:
            phrase = cut_phrase(phrase)
            if change > 0:
                if hot < top_n:
                    data_h[phrase] = change
                    text_h.append('+' + str(("%.1f" % change)) + '%')
                    hot += 1
            elif cold < top_n:
                data_c[phrase] = change
                text_c.append(str(("%.1f" % change)) + '%')
                cold += 1
    if len(data_h.keys()) > 0:
        phrases_h = sorted(data_h, key=data_h.get)
        changes_h = sorted(data_h.values())
        source = ColumnDataSource(
            dict(y=phrases_h, x=changes_h, text=text_h[::-1]))
        title = header_h
        plot_h = figure(plot_width=600, plot_height=400, tools="save", title=title,
                        x_range=ranges.Range1d(start=0, end=max(changes_h) + 20),
                        y_range=source.data["y"], tooltips='@y<br>change: @x')
        labels = LabelSet(x='x', y='y', text='text', level='glyph',
                          x_offset=5, y_offset=0, source=source,
                          render_mode='canvas', text_color="#1d6d34")
        plot_h.hbar(source=source, right='x', y='y', left=0, height=0.5, color="#1d6d34")
        plot_h.add_layout(labels)
        plot_h.xaxis.visible = False
    if len(data_c.keys()) > 0:
        phrases_c = sorted(data_c, key=data_c.get)
        changes_c = sorted(data_c.values())
        source = ColumnDataSource(
            dict(y=phrases_c[::-1], x=changes_c[::-1], text=text_c[::-1]))
        plot_c = figure(plot_width=600, plot_height=400, tools="save",
                        title=header_c,
                        x_range=ranges.Range1d(start=min(changes_c) - 10, end=20),
                        y_range=source.data["y"],
                        tooltips='@y<br>change: @x')
        labels = LabelSet(x=0, y='y', text='text', level='glyph',
                          x_offset=5, y_offset=0, source=source,
                          render_mode='canvas', text_color="#912605")
        plot_c.hbar(source=source, right='x', y='y', left=0, height=0.5,
                    color="#912605")
        plot_c.add_layout(labels)
        plot_c.xaxis.visible = False

    return [plot_h, plot_c]


def create_trend_clustering_graph(top_n):
    if 'Trend Clustering' not in gd['reports'].values:  # no clustering data
        return None
    header = gd['reports'][5]
    scatter_series_pos = gd[(gd['x_tre'] != -1) & (gd['x_tre'].notna()) & (gd['change'] > 0)]
    scatter_series_neg = gd[
        (gd['x_tre'] != -1) & (gd['x_tre'].notna()) & (gd['change'] < 0)]
    hot_phrases = scatter_series_pos['trends'].values
    cold_phrases = scatter_series_neg['trends'].values
    hot_change = scatter_series_pos['change'].values
    cold_change = scatter_series_neg['change'].values
    hot_x_t = scatter_series_pos['x_tre'].values
    cold_x_t = scatter_series_neg['x_tre'].values
    hot_y_t = scatter_series_pos['y_tre'].values
    cold_y_t = scatter_series_neg['y_tre'].values
    phrases_h, changes_h, x_h, y_h, hover_data_h = [], [], [], [], []
    phrases_c, changes_c, x_c, y_c, hover_data_c = [], [], [], [], []
    opacity_h, opacity_c = [], []
    hot, cold = 0, 0  # counters for each list
    if len(hot_change) > 0 and len(cold_change) > 0:
        highest_change = max(hot_change[0], abs(cold_change[0]))
        opacity_norm = 1 / (highest_change) if highest_change != 0 else 0.05

    for i in range(len(hot_phrases)):
        if hot == top_n:
            break
        phrase = hot_phrases[i]
        change = float(hot_change[i])
        if phrase in white_list:
            phrases_h.append(phrase)
            changes_h.append(change)
            x_h.append(hot_x_t[i])
            y_h.append(hot_y_t[i])
            hover_data_h.append(phrase)
            opacity = opacity_norm * abs(change)
            opacity_h.append(
                opacity if opacity > 0.05 else 0.05)
            hot += 1

    for j in range(len(cold_phrases)):
        if cold == top_n:
            break
        phrase = cold_phrases[j]
        change = float(cold_change[j])
        if phrase in white_list:
            phrases_c.append(phrase)
            changes_c.append(change)
            x_c.append(cold_x_t[j])
            y_c.append(cold_y_t[j])
            hover_data_c.append(phrase)
            opacity = opacity_norm * abs(change)
            opacity_c.append(
                opacity if opacity > 0.05 else 0.05)
            cold += 1

    opacities = []
    colors = []
    # hot:
    for i in range(len(x_h)):
        opacities.append(opacity_h[i])
        colors.append('#4c823d')
    # cold:
    for i in range(len(x_c)):
        opacities.append(opacity_c[i])
        colors.append('#d6400a')

    source = ColumnDataSource(data=dict(
        x=x_h + x_c,
        y=y_h + y_c,
        topic=hover_data_h + hover_data_c,
        opacities=opacities,
        colors=colors
    ))
    p = figure(title=header, plot_width=600, plot_height=400, tooltips='@topic')
    p.circle(x='x', y='y', size=13, color='colors',
             alpha='opacities', source=source)
    return p


def create_top_before_graph(top_n):
    global ref_data_name
    header = gd['reports'][0]
    if ref_data_name == "":
        ref_data_name = header.split('(')[1][:-1]
    phrases = gd.values[:, 1]
    importance = gd.values[:, 2]
    valid_phrases, valid_imp = [], []
    ctr = 0
    for i in range(len(phrases)):
        if ctr == top_n:
            break
        p = phrases[i]
        if p in white_list:
            p = cut_phrase(p)
            valid_phrases.append(p)
            valid_imp.append(importance[i])
            ctr += 1

    source = ColumnDataSource(dict(y=valid_phrases[::-1], x=valid_imp[::-1]))
    x_label = 'Importance'
    plot = figure(plot_width=600, plot_height=400, tools="save",
                  x_axis_label=x_label,
                  title=header,
                  y_range=source.data["y"],
                  tooltips='@y<br>@x')
    plot.hbar(source=source, right='x', y='y', left=0, height=0.5,
              color="#335fa5")
    return plot


def create_top_after_graph(topN):
    global target_data_name
    header = gd['reports'][1]
    if target_data_name == "":
        target_data_name = header.split('(')[1][:-1]
    phrases = gd.values[:, 5]
    importance = gd.values[:, 6]
    valid_phrases, valid_imp = [], []
    ctr = 0
    for i in range(len(phrases)):
        if ctr == topN:
            break
        p = phrases[i]
        if p in white_list:
            p = cut_phrase(p)
            valid_phrases.append(p)
            valid_imp.append(importance[i])
            ctr += 1

    source = ColumnDataSource(
        dict(y=valid_phrases[::-1], x=valid_imp[::-1]))
    x_label = 'Importance'
    plot = figure(plot_width=600, plot_height=400, tools="save",
                  x_axis_label=x_label,
                  title=header,
                  y_range=source.data["y"],
                  tooltips='@y<br>@x')
    plot.hbar(source=source, right='x', y='y', left=0, height=0.5,
              color="#335fa5")
    return plot


def handle_selected_graph(selected):
    logger.info("handle selected graph. selected = %s", str(selected))
    label.text = "Please Wait..."
    global apply_filter
    if top_before is not None:
        if apply_filter:
            regenerate_graphs(int(top_n_dropdown.value), int(top_n_clusters_dropdown.value))
            apply_filter = False
        if selected == 0:
            if trend_clustering is None:
                graphs_area.children = [Row(top_before, top_after),
                                        Row(hot_trends, cold_trends),
                                        Row(custom_trends)
                                        ]
            else:
                graphs_area.children = [Row(top_before, top_after),
                                        Row(hot_trends, cold_trends),
                                        Row(trend_clustering, custom_trends),
                                        Row(topic_clustering_before, topic_clustering_after)
                                        ]
        if selected == 1:
            graphs_area.children = [Row(top_before, top_after)]
        if selected == 2:
            graphs_area.children = [Row(hot_trends, cold_trends)]
        if selected == 3:
            if trend_clustering is None:
                graphs_area.children = [Div(text='no word embedding data')]
            else:
                graphs_area.children = [trend_clustering]
        if selected == 4:
            graphs_area.children = [custom_trends]
        if selected == 5:
            if topic_clustering_before is None:
                graphs_area.children = [Div(text='no word embedding data')]
            else:
                graphs_area.children = [Row(topic_clustering_before,
                                        topic_clustering_after)]
        if selected == 6:
            filter_topics_table_source.data = {'topics': filter_rows}
            filter_topics_table_source.selected.indices = get_valid_indices()
            filter_custom_table_source.data = {'topics': filter_rows}
            filter_custom_table_source.selected.indices = get_custom_indices()
            refresh_filter_area()
            apply_filter = False
    label.text = ""


def draw_ui(top_n_phrases, top_n_clusters, active_area):
    logger.info("draw ui. top_n_phrases=%s, top_n_clusters=%s",
                str(top_n_phrases), str(top_n_clusters))
    create_graphs(top_n_phrases, top_n_clusters)
    handle_selected_graph(active_area)


# define callbacks
# pylint: disable=unused-argument
def top_n_phrases_changed_callback(value, old, new):
    logger.info("top n changed to: %s", str(new))
    top_n_dropdown.label = "Show Top " + str(new) + " Phrases"
    draw_ui(new, top_n_clusters_dropdown.value, radio_group_area.active)
    # create_graphs(new)
    # handle_selected_graph(radio_group_area.active)


# pylint: disable=unused-argument
def top_n_clusters_changed_callback(value, old, new):
    global dont_regenerate
    logger.info("top n clusters changed to: %s", str(new))
    top_n_clusters_dropdown.label = "Show Top " + str(new) + " Clusters"
    if not dont_regenerate:
        draw_ui(top_n_dropdown.value, new, radio_group_area.active)
    else:
        logger.info("skip draw_ui")
        dont_regenerate = False


# pylint: disable=unused-argument
def selected_graph_changed_callback(active, old, new):
    global filter_item
    logger.info("selected graph changed to: %s", str(new))
    reset_filters()
    handle_selected_graph(new)


# pylint: disable=unused-argument
def filter_topic_selected_callback(indices, old, new):
    # update all_topics according to new selected items
    logger.info("filter topic callback")
    filter_label.text = "Please Wait..."
    global all_topics, apply_filter
    if filter_item != -1:  # on filter/search state there are None values so need
        # to separate this case
        apply_filter = True
        if filter_item in new:
            all_topics[filter_item][filter_tabs.active + 1] = '1'
        else:
            all_topics[filter_item][filter_tabs.active + 1] = '0'
    elif new != [-1]:
        apply_filter = True
        selected_topics = [filter_topics_table_source.data['topics'][x] for x in new]
        for i, line in enumerate(all_topics):
            if line[0] in selected_topics:
                all_topics[i][1] = '1'
            else:
                all_topics[i][1] = '0'
    filter_label.text = ""


# pylint: disable=unused-argument
def filter_custom_selected_callback(indices, old, new):
    logger.info("filter custom callback")
    filter_label.text = "Please Wait..."
    global all_topics, apply_filter
    if new != [-1]:
        apply_filter = True
        selected_topics = [
            filter_custom_table_source.data['topics'][x] for x in
            new]
        for i, line in enumerate(all_topics):
            if line[0] in selected_topics:
                all_topics[i][2] = '1'
            else:
                all_topics[i][2] = '0'
    filter_label.text = ""


def search_topic_callback():
    global filter_item
    logger.info("search topic callback")
    filter_label.text = ""
    text_to_search = search_input_box.value
    if len(text_to_search) > 0:
        x = search(text_to_search)
        logger.info("x=%s", str(x))
        if x != -1:
            reset_filters()
            filter_item = x
            if filter_tabs.active == 0:
                logger.info("is index %s a valid topic: %s", str(x), str(is_valid_topic(x)))
                filter_topics_table.view.filters[0].indices = [x]
                filter_topics_table_source.data = {'topics': filter_rows}  # just for
                # refreshing the ui
            else:
                filter_custom_table.view.filters[0].indices = [x]
                filter_custom_table_source.data = {
                    'topics': filter_rows}
        else:
            filter_label.text = "no results"
            logger.info("no results")
    refresh_filter_area()


def tab_changed_callback(_):
    reset_filters()


def clear_search_callback():
    reset_filters()
    search_input_box.value = ""
    refresh_filter_area()


def re_analyze_callback():
    re_analyze()


# set callbacks

top_n_dropdown.on_change('value', top_n_phrases_changed_callback)
top_n_clusters_dropdown.on_change('value', top_n_clusters_changed_callback)
radio_group_area.on_change('active', selected_graph_changed_callback)
# pylint: disable=no-member
filter_topics_table_source.selected.on_change('indices', filter_topic_selected_callback)
filter_custom_table_source.selected.on_change('indices', filter_custom_selected_callback)
search_button.on_click(search_topic_callback)
clear_button.on_click(clear_search_callback)
filter_tabs.on_change('active', tab_changed_callback)
analyse_button.on_click(re_analyze_callback)


# start app

draw_ui(top_n_dropdown.value, top_n_clusters_dropdown.value, radio_group_area.active)

doc = curdoc()
main_title = "Trend Analysis"
doc.title = main_title
doc.add_root(grid)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf


# taken from keras previous versions: https://github.com/keras-team/keras/issues/5400
def precision_score(y_true, y_pred):
    """Precision metric.

    Only computes a batch-wise average of precision.

    Computes the precision, a metric for multi-label classification of
    how many selected items are relevant.
    """
    K = tf.keras.backend
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision


def recall_score(y_true, y_pred):
    """Recall metric.

    Only computes a batch-wise average of recall.

    Computes the recall, a metric for multi-label classification of
    how many relevant items are selected.
    """
    K = tf.keras.backend
    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall


def f1(y_true, y_pred):
    """

    Args:
        y_true:
        y_pred:

    Returns:

    """
    K = tf.keras.backend
    precision = precision_score(y_true, y_pred)
    recall = recall_score(y_true, y_pred)
    return 2 * ((precision * recall) / (precision + recall + K.epsilon()))


class NpSemanticSegClassifier:
    """
    NP Semantic Segmentation classifier model (based on tf.Keras framework).

    Args:
        num_epochs(int): number of epochs to train the model
        **callback_args (dict): callback args keyword arguments to init a Callback for the model
        loss: the model's cost function. Default is 'tf.keras.losses.binary_crossentropy' loss
        optimizer (:obj:`tf.keras.optimizers`): the model's optimizer. Default is 'adam'
    """

    def __init__(self, num_epochs, callback_args, loss='binary_crossentropy', optimizer='adam',
                 batch_size=128, ):
        """
        Args:
            num_epochs(int): number of epochs to train the model
            callback_args (dict): callback args keyword arguments to init Callback for the model
            loss: the model's loss function. Default is 'tf.keras.losses.binary_crossentropy' loss
            optimizer (:obj:`tf.keras.optimizers`): the model's optimizer. Default is `adam`
            batch_size (int):  batch size
        """
        self.model = None
        self.loss = loss
        self.optimizer = optimizer
        self.epochs = num_epochs
        self.callback_args = callback_args
        self.batch_size = batch_size

    def build(self, input_dim):
        """
        Build the model's layers
        Args:
            input_dim (int): the first layer's input_dim
        """
        first_layer_dens = 64
        second_layer_dens = 64
        output_layer_dens = 1
        model = tf.keras.models.Sequential()
        model.add(tf.keras.layers.Dense(first_layer_dens, activation='relu', input_dim=input_dim))
        model.add(tf.keras.layers.Dropout(0.5))
        model.add(tf.keras.layers.Dense(second_layer_dens, activation='relu'))
        model.add(tf.keras.layers.Dropout(0.5))
        model.add(tf.keras.layers.Dense(output_layer_dens, activation='sigmoid'))
        metrics = ['binary_accuracy', precision_score, recall_score, f1]
        # Compile model
        model.compile(loss=self.loss, optimizer=self.optimizer, metrics=metrics)
        self.model = model

    def fit(self, train_set):
        """
        Train and fit the model on the datasets

        Args:
            train_set (:obj:`numpy.ndarray`): The train set
            args: callback_args and epochs from ArgParser input
        """
        self.model.fit(train_set['X'], train_set['y'], epochs=self.epochs,
                       batch_size=self.batch_size, verbose=2)

    def save(self, model_path):
        """
        Save the model's prm file in model_path location

        Args:
            model_path(str): local path for saving the model
        """
        # serialize model to JSON
        model_json = self.model.to_json()
        with open(model_path[:-2] + 'json', 'w') as json_file:
            json_file.write(model_json)
        # serialize weights to HDF5
        self.model.save_weights(model_path)
        print("Saved model to disk")

    def load(self, model_path):
        """
        Load pre-trained model's .h5 file to NpSemanticSegClassifier object

        Args:
            model_path(str): local path for loading the model
        """
        # load json and create model
        with open(model_path[:-2] + 'json', 'r') as json_file:
            loaded_model_json = json_file.read()
        loaded_model = tf.keras.models.model_from_json(loaded_model_json)
        # load weights into new model
        loaded_model.load_weights(model_path)
        print("Loaded model from disk")
        self.model = loaded_model

    def eval(self, test_set):
        """
        Evaluate the model's test_set on error_rate, test_accuracy_rate and precision_recall_rate

        Args:
            test_set (:obj:`numpy.ndarray`): The test set

        Returns:
            tuple(float): loss, binary_accuracy, precision, recall and f1 measures
        """
        return self.model.evaluate(test_set['X'], test_set['y'], batch_size=128, verbose=2)

    def get_outputs(self, test_set):
        """
        Classify the dataset on the model

        Args:
            test_set (:obj:`numpy.ndarray`): The test set

        Returns:
            list(:obj:`numpy.ndarray`): model's predictions
        """
        return self.model.predict(test_set)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import logging
import os
import pickle
from os.path import join

from nlp_architect.common.cdc.mention_data import MentionData
from nlp_architect.data.cdc_resources.embedding.embed_elmo import ElmoEmbedding
from nlp_architect.utils import io

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


def load_elmo_for_vocab(mentions):
    """
    Create the embedding using the cache logic in the embedding class
    Args:
        mentions:

    Returns:

    """
    elmo_embeddings = ElmoEmbedding()

    for mention in mentions:
        elmo_embeddings.get_head_feature_vector(mention)

    logger.info('Total words/contexts in vocabulary %d', len(elmo_embeddings.cache))
    return elmo_embeddings.cache


def elmo_dump():
    out_file = args.output
    mention_files = list()
    if os.path.isdir(args.mentions):
        for (dirpath, _, files) in os.walk(args.mentions):
            for file in files:
                if file == '.DS_Store':
                    continue

                mention_files.append(join(dirpath, file))
    else:
        mention_files.append(args.mentions)

    mentions = []
    for _file in mention_files:
        mentions.extend(MentionData.read_mentions_json_to_mentions_data_list(_file))

    elmo_ecb_embeddings = load_elmo_for_vocab(mentions)

    with open(out_file, 'wb') as f:
        pickle.dump(elmo_ecb_embeddings, f)

    logger.info('Saving dump to file-%s', out_file)


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Create Elmo Embedding dataset only dump')
    parser.add_argument('--mentions', type=str, help='mentions_file file', required=True)
    parser.add_argument('--output', type=str, help='location were to create dump file',
                        required=True)

    args = parser.parse_args()

    if os.path.isdir(args.mentions):
        io.validate_existing_directory(args.mentions)
    else:
        io.validate_existing_filepath(args.mentions)

    elmo_dump()
    print('Done!')



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import unicode_literals, print_function, division, \
    absolute_import
import json


class HighLevelDoc:
    """
    object for annotation documents

    Args:
        self.doc_text (str): document text
        self.annotation_set (list(str)): list of all annotations in doc
        self.spans (list(dict)): list of span dict, each span_dict is structured as follows:
            { 'end': (int), 'start': (int), 'type': (str) string of annotation }
    """
    def __init__(self):
        self.doc_text = None
        self.annotation_set = []
        self.spans = []

    def json(self):
        """
        Return json representations of the object

        Returns:
            :obj:`json`: json representations of the object
        """
        return json.dumps(self.__dict__)

    def pretty_json(self):
        """
        Return pretty json representations of the object

        Returns:
            :obj:`json`: pretty json representations of the object
        """
        return json.dumps(self.__dict__, indent=4)

    def displacy_doc(self):  # only change annotations to lowercase
        """
        Return doc adapted to displacyENT expected input
        """
        self.annotation_set = [annotation.lower() for annotation in self.annotation_set]
        return self.__dict__



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import json


def merge_punct_tok(merged_punct_sentence, last_merged_punct_index, punct_text, is_traverse):
    # merge the text of the punct tok
    if is_traverse:
        merged_punct_sentence[last_merged_punct_index]["text"] = \
            punct_text + merged_punct_sentence[last_merged_punct_index]["text"]
    else:
        merged_punct_sentence[last_merged_punct_index]["text"] = \
            merged_punct_sentence[last_merged_punct_index]["text"] + punct_text


def find_correct_index(orig_gov, merged_punct_sentence):
    for tok_index, tok in enumerate(merged_punct_sentence):
        if tok["start"] == orig_gov["start"] and tok["len"] == orig_gov["len"] and tok["pos"] == \
                orig_gov["pos"] and tok["text"] == orig_gov["text"]:
            return tok_index
    return None


def fix_gov_indexes(merged_punct_sentence, sentence):
    for merged_token in merged_punct_sentence:
        tok_gov = merged_token['gov']
        if tok_gov == -1:  # gov is root
            merged_token['gov'] = -1
        else:
            orig_gov = sentence[tok_gov]
            correct_index = find_correct_index(orig_gov, merged_punct_sentence)
            merged_token['gov'] = correct_index


def merge_punctuation(sentence):
    merged_punct_sentence = []
    tmp_punct_text = None
    punct_text = None
    last_merged_punct_index = -1
    for tok_index, token in enumerate(sentence):
        if token['rel'] == 'punct':
            punct_text = token["text"]
            if tok_index < 1:  # this is the first tok - append to the next token
                tmp_punct_text = punct_text
            else:  # append to the previous token
                merge_punct_tok(merged_punct_sentence, last_merged_punct_index, punct_text,
                                False)
        else:
            merged_punct_sentence.append(token)
            last_merged_punct_index = last_merged_punct_index + 1
            if tmp_punct_text is not None:
                merge_punct_tok(merged_punct_sentence, last_merged_punct_index, punct_text,
                                True)
                tmp_punct_text = None
    return merged_punct_sentence


class CoreNLPDoc(object):
    """Object for core-components (POS, Dependency Relations, etc).

    Attributes:
        _doc_text: the doc text
        _sentences: list of sentences, each word in a sentence is
            represented by a dictionary, structured as follows: {'start': (int), 'len': (int),
            'pos': (str), 'ner': (str), 'lemma': (str), 'gov': (int), 'rel': (str)}
    """
    def __init__(self, doc_text: str = '', sentences: list = None):
        if sentences is None:
            sentences = []
        self._doc_text = doc_text
        self._sentences = sentences

    @property
    def doc_text(self):
        return self._doc_text

    @doc_text.setter
    def doc_text(self, val):
        self._doc_text = val

    @property
    def sentences(self):
        return self._sentences

    @sentences.setter
    def sentences(self, val):
        self._sentences = val

    @staticmethod
    def decoder(obj):
        if '_doc_text' in obj and '_sentences' in obj:
            return CoreNLPDoc(obj['_doc_text'], obj['_sentences'])
        return obj

    def __repr__(self):
        return self.pretty_json()

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        return self.sentences.__iter__()

    def __len__(self):
        return len(self.sentences)

    def json(self):
        """Returns json representations of the object."""
        return json.dumps(self.__dict__)

    def pretty_json(self):
        """Returns pretty json representations of the object."""
        return json.dumps(self.__dict__, indent=4)

    def sent_text(self, i):
        parsed_sent = self.sentences[i]
        first_tok, last_tok = parsed_sent[0], parsed_sent[-1]
        return self.doc_text[first_tok['start']: last_tok['start'] + last_tok['len']]

    def sent_iter(self):
        for parsed_sent in self.sentences:
            first_tok, last_tok = parsed_sent[0], parsed_sent[-1]
            sent_text = self.doc_text[first_tok['start']: last_tok['start'] + last_tok['len']]
            yield sent_text, parsed_sent

    def brat_doc(self):
        """Returns doc adapted to BRAT expected input."""
        doc = {'text': '', 'entities': [], 'relations': []}
        tok_count = 0
        rel_count = 1
        for sentence in self.sentences:
            sentence_start = sentence[0]['start']
            sentence_end = sentence[-1]['start'] + sentence[-1]['len']
            doc['text'] = doc['text'] + '\n' + self.doc_text[sentence_start:sentence_end]
            token_offset = tok_count

            for token in sentence:
                start = token['start']
                end = start + token['len']
                doc['entities'].append(['T' + str(tok_count), token['pos'], [[start, end]]])

                if token['gov'] != -1 and token['rel'] != 'punct':
                    doc['relations'].append(
                        [rel_count, token['rel'], [['', 'T' + str(token_offset + token['gov'])],
                                                   ['', 'T' + str(tok_count)]]])
                    rel_count += 1
                tok_count += 1
        doc['text'] = doc['text'][1:]
        return doc

    def displacy_doc(self):
        """Return doc adapted to displacyENT expected input."""
        doc = []
        for sentence in self.sentences:
            sentence_doc = {'arcs': [], 'words': []}
            # Merge punctuation:
            merged_punct_sentence = merge_punctuation(sentence)
            fix_gov_indexes(merged_punct_sentence, sentence)
            for tok_index, token in enumerate(merged_punct_sentence):
                sentence_doc['words'].append({'text': token['text'], 'tag': token['pos']})
                dep_tok = tok_index
                gov_tok = token['gov']
                direction = 'left'
                arc_start = dep_tok
                arc_end = gov_tok
                if dep_tok > gov_tok:
                    direction = 'right'
                    arc_start = gov_tok
                    arc_end = dep_tok
                if token['gov'] != -1 and token['rel'] != 'punct':
                    sentence_doc['arcs'].append({'dir': direction, 'label': token['rel'],
                                                 'start': arc_start, 'end': arc_end})
            doc.append(sentence_doc)
        return doc



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import socket
from os import path
from subprocess import run

SOLUTIONS_PATH = path.dirname(path.realpath(__file__))

solution_uis = {
    'set_expansion': path.join(SOLUTIONS_PATH, 'set_expansion', 'ui'),
    'trend_analysis': path.join(SOLUTIONS_PATH, 'trend_analysis', 'ui')
}


def check_if_ip(ip_str):
    try:
        socket.inet_aton(ip_str)
        return True
    except socket.error:
        return False


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--solution', type=str, choices=['set_expansion', 'trend_analysis'],
                        help='Solution UI to initialize')
    parser.add_argument('--address', type=str, default='',
                        help='IP address to use for UI server ')
    parser.add_argument('--port', type=int, default=1010,
                        help='Port number')
    args = parser.parse_args()
    if args.address and not check_if_ip(args.address):
        print('given address is not in a valid ip address format')
        exit(1)

    cmd_str = 'bokeh serve --show {} --port {}'.format(solution_uis[args.solution], args.port)
    if args.address:
        cmd_str += ' --address={} ' \
                   '--allow-websocket-origin={}:{}'.\
            format(args.port, args.address, args.port)

    run(cmd_str)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from typing import List

from nlp_architect.common.cdc.mention_data import MentionData


class Cluster(object):
    def __init__(self, coref_chain: int = -1) -> None:
        """
        Object represent a set of mentions with same coref chain id

        Args:
            coref_chain (int): the cluster id/coref_chain value
        """
        self.mentions = []
        self.cluster_strings = []
        self.merged = False
        self.coref_chain = coref_chain
        self.mentions_corefs = set()

    def get_mentions(self):
        return self.mentions

    def add_mention(self, mention: MentionData) -> None:
        if mention is not None:
            mention.predicted_coref_chain = self.coref_chain
            self.mentions.append(mention)
            self.cluster_strings.append(mention.tokens_str)
            self.mentions_corefs.add(mention.coref_chain)

    def merge_clusters(self, cluster) -> None:
        """
        Args:
            cluster: cluster to merge this cluster with
        """
        for mention in cluster.mentions:
            mention.predicted_coref_chain = self.coref_chain

        self.mentions.extend(cluster.mentions)
        self.cluster_strings.extend(cluster.cluster_strings)
        self.mentions_corefs.update(cluster.mentions_corefs)

    def get_cluster_id(self) -> str:
        """
        Returns:
            A generated cluster unique Id created from cluster mentions ids
        """
        return '$'.join([mention.mention_id for mention in self.mentions])


class Clusters(object):
    cluster_coref_chain = 1000

    def __init__(self, topic_id: str, mentions: List[MentionData] = None) -> None:
        """

        Args:
            mentions: ``list[MentionData]``, required
                The initial mentions to create the clusters from
        """
        self.clusters_list = []
        self.topic_id = topic_id
        self.set_initial_clusters(mentions)

    def set_initial_clusters(self, mentions: List[MentionData]) -> None:
        """

        Args:
            mentions: ``list[MentionData]``, required
                The initial mentions to create the clusters from

        """
        if mentions:
            for mention in mentions:
                cluster = Cluster(Clusters.cluster_coref_chain)
                cluster.add_mention(mention)
                self.clusters_list.append(cluster)
                Clusters.cluster_coref_chain += 1

    def clean_clusters(self) -> None:
        """
        Remove all clusters that were already merged with other clusters
        """

        self.clusters_list = [cluster for cluster in self.clusters_list if not cluster.merged]

    def set_coref_chain_to_mentions(self) -> None:
        """
        Give all cluster mentions the same coref ID as cluster coref chain ID

        """
        for cluster in self.clusters_list:
            for mention in cluster.mentions:
                mention.predicted_coref_chain = str(cluster.coref_chain)

    def add_cluster(self, cluster: Cluster) -> None:
        self.clusters_list.append(cluster)

    def add_clusters(self, clusters) -> None:
        for cluster in clusters.clusters_list:
            self.clusters_list.append(cluster)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import os
from typing import Dict, List, Set

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import \
    OnlineOROfflineMethod, RelationType
from nlp_architect.utils.io import load_json_file
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


class ReferentDictRelationExtraction(RelationExtraction):
    def __init__(self, method: OnlineOROfflineMethod = OnlineOROfflineMethod.ONLINE,
                 ref_dict: str = None):
        """
        Extract Relation between two mentions according to Referent Dictionary knowledge

        Args:
            method (optional): OnlineOROfflineMethod.{ONLINE/OFFLINE} run against full referent
                dictionary or a sub-set of (default = ONLINE)
            ref_dict (required): str Location of referent dictionary file to work with
        """
        logger.info('Loading ReferentDict module')
        if ref_dict is not None and os.path.isfile(ref_dict):
            if method == OnlineOROfflineMethod.OFFLINE:
                self.ref_dict = load_json_file(ref_dict)
            elif method == OnlineOROfflineMethod.ONLINE:
                self.ref_dict = self.load_reference_dict(ref_dict)
            logger.info('ReferentDict module lead successfully')
        else:
            raise FileNotFoundError('Referent Dict file not found or not in path:' + ref_dict)

        super(ReferentDictRelationExtraction, self).__init__()

    def extract_all_relations(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> Set[RelationType]:
        ret_ = set()
        ret_.add(self.extract_sub_relations(mention_x, mention_y, RelationType.REFERENT_DICT))
        return ret_

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        if relation is not RelationType.REFERENT_DICT:
            return RelationType.NO_RELATION_FOUND

        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str
        if StringUtils.is_pronoun(mention_x_str.lower()) or StringUtils.is_pronoun(
                mention_y_str.lower()):
            return RelationType.NO_RELATION_FOUND

        if self.is_referent_dict(mention_x, mention_y):
            return RelationType.REFERENT_DICT

        return RelationType.NO_RELATION_FOUND

    def is_referent_dict(self, mention_x: MentionDataLight, mention_y: MentionDataLight) -> bool:
        """
        Check if input mentions has referent dictionary relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            bool
        """
        match_result = False
        x_head = mention_x.mention_head
        y_head = mention_y.mention_head
        x_head_lemma = mention_x.mention_head_lemma
        y_head_lemma = mention_y.mention_head_lemma

        if (x_head in self.ref_dict and y_head in self.ref_dict[x_head]) or (
                y_head in self.ref_dict and x_head in self.ref_dict[y_head]):
            match_result = True
        if (x_head_lemma in self.ref_dict and y_head_lemma in self.ref_dict[x_head_lemma]) or \
                (y_head_lemma in self.ref_dict and x_head_lemma in self.ref_dict[y_head_lemma]):
            match_result = True
        if (x_head_lemma in self.ref_dict and y_head in self.ref_dict[x_head_lemma]) or (
                y_head_lemma in self.ref_dict and x_head in self.ref_dict[y_head_lemma]):
            match_result = True
        if (y_head in self.ref_dict and x_head_lemma in self.ref_dict[y_head]) or \
                (x_head in self.ref_dict and y_head_lemma in self.ref_dict[x_head]):
            match_result = True

        return match_result

    @staticmethod
    def get_supported_relations():
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.REFERENT_DICT]

    @staticmethod
    def load_reference_dict(dict_fname: str) -> Dict[str, List[str]]:
        """
        Method to load referent dictionary to memory

        Returns:
            List[RelationType]
        """
        word_dict = {}
        first = True
        with open(dict_fname, 'r', encoding="utf-8") as f:
            for line in f:
                if first:
                    first = False
                    continue
                word1, word2, _, npmi = line.strip().split('\t')
                npmi = float(npmi)
                if npmi >= 0.2:
                    if word1 not in word_dict:
                        word_dict[word1] = []
                    word_dict[word1].append(word2)
        return word_dict



#!/usr/bin/env python

# This file was taken from:
# https://github.com/Kronuz/ansi2html/blob/master/ansi2html.py
# Copyright © 2013 German M. Bravo (Kronuz)
# License: https://github.com/Kronuz/ansi2html/blob/master/LICENSE
# Changes made to original file: removed main, added run() function.
from __future__ import print_function

import re
import sys

_ANSI2HTML_STYLES = {}
# pylint: disable=anomalous-backslash-in-string
ANSI2HTML_CODES_RE = re.compile('(?:\033\\[(\\d+(?:;\\d+)*)?([cnRhlABCDfsurgKJipm]))')
ANSI2HTML_PALETTE = {
    # See http://ethanschoonover.com/solarized
    'solarized': ['#073642', '#D30102', '#859900', '#B58900', '#268BD2', '#D33682', '#2AA198',
                  '#EEE8D5', '#002B36', '#CB4B16', '#586E75', '#657B83', '#839496', '#6C71C4',
                  '#93A1A1', '#FDF6E3'],
    # Above mapped onto the xterm 256 color palette
    'solarized-xterm': ['#262626', '#AF0000', '#5F8700', '#AF8700', '#0087FF', '#AF005F',
                        '#00AFAF', '#E4E4E4', '#1C1C1C', '#D75F00', '#585858', '#626262',
                        '#808080', '#5F5FAF', '#8A8A8A', '#FFFFD7'],
    # Gnome default:
    'tango': ['#000000', '#CC0000', '#4E9A06', '#C4A000', '#3465A4', '#75507B', '#06989A',
              '#D3D7CF', '#555753', '#EF2929', '#8AE234', '#FCE94F', '#729FCF', '#AD7FA8',
              '#34E2E2', '#EEEEEC'],
    # xterm:
    'xterm': ['#000000', '#CD0000', '#00CD00', '#CDCD00', '#0000EE', '#CD00CD', '#00CDCD',
              '#E5E5E5', '#7F7F7F', '#FF0000', '#00FF00', '#FFFF00', '#5C5CFF', '#FF00FF',
              '#00FFFF', '#FFFFFF'],
    'console': ['#000000', '#AA0000', '#00AA00', '#AA5500', '#0000AA', '#AA00AA', '#00AAAA',
                '#AAAAAA', '#555555', '#FF5555', '#55FF55', '#FFFF55', '#5555FF', '#FF55FF',
                '#55FFFF', '#FFFFFF'],
}


def _ansi2html_get_styles(palette):
    if palette not in _ANSI2HTML_STYLES:
        p = ANSI2HTML_PALETTE.get(palette, ANSI2HTML_PALETTE['console'])

        regular_style = {
            '1': '',  # bold
            '2': 'opacity:0.5',
            '4': 'text-decoration:underline',
            '5': 'font-weight:bold',
            '7': '',
            '8': 'display:none',
        }
        bold_style = regular_style.copy()
        for i in range(8):
            regular_style['3%s' % i] = 'color:%s' % p[i]
            regular_style['4%s' % i] = 'background-color:%s' % p[i]

            bold_style['3%s' % i] = 'color:%s' % p[i + 8]
            bold_style['4%s' % i] = 'background-color:%s' % p[i + 8]

        # The default xterm 256 colour p:
        indexed_style = {}
        for i in range(16):
            indexed_style['%s' % i] = p[i]

        for rr in range(6):
            for gg in range(6):
                for bb in range(6):
                    i = 16 + rr * 36 + gg * 6 + bb
                    r = (rr * 40 + 55) if rr else 0
                    g = (gg * 40 + 55) if gg else 0
                    b = (bb * 40 + 55) if bb else 0
                    indexed_style['%s' % i] = ''.join(
                        '%02X' % c if 0 <= c <= 255 else None for c in (r, g, b))

        for g in range(24):
            i = g + 232
            j = g * 10 + 8
            indexed_style['%s' % i] = ''.join(
                '%02X' % c if 0 <= c <= 255 else None for c in (j, j, j))

        _ANSI2HTML_STYLES[palette] = (regular_style, bold_style, indexed_style)
    return _ANSI2HTML_STYLES[palette]


def ansi2html(text, palette='solarized'):
    # pylint: disable=too-many-branches, too-many-statements
    def _ansi2html(m):
        if m.group(2) != 'm':
            return ''
        state = None
        sub = ''
        cs = m.group(1)
        cs = cs.strip() if cs else ''
        for c in cs.split(';'):
            c = c.strip().lstrip('0') or '0'
            if c == '0':
                while stack:
                    sub += '</span>'
                    stack.pop()
            elif c in ('38', '48'):
                extra = [c]
                state = 'extra'
            elif state == 'extra':
                if c == '5':
                    state = 'idx'
                elif c == '2':
                    state = 'r'
            elif state:
                if state == 'idx':
                    extra.append(c)
                    state = None
                    color = indexed_style.get(c)  # 256 colors
                    if color is not None:
                        sub += \
                            '<span style="%s:%s">' % \
                            ('color' if extra[0] == '38' else 'background-color', color)
                        stack.append(extra)
                elif state in ('r', 'g', 'b'):
                    extra.append(c)
                    if state == 'r':
                        state = 'g'
                    elif state == 'g':
                        state = 'b'
                    else:
                        state = None
                        try:
                            color = '#' + ''.join(
                                '%02X' % c if 0 <= c <= 255 else None for x in extra for c in
                                [int(x)])
                        except (ValueError, TypeError):
                            pass
                        else:
                            sub += \
                                '<span style="%s:%s">' % \
                                ('color' if extra[0] == '38' else 'background-color', color)
                            stack.append(extra)
            else:
                if '1' in stack:
                    style = bold_style.get(c)
                else:
                    style = regular_style.get(c)
                if style is not None:
                    sub += '<span style="%s">' % style
                    stack.append(c)  # Still needs to be added to the stack even if style is empty
                    #  (so it can check '1' in stack above, for example)
        return sub
    stack = []
    regular_style, bold_style, indexed_style = _ansi2html_get_styles(palette)
    sub = ANSI2HTML_CODES_RE.sub(_ansi2html, text)
    while stack:
        sub += '</span>'
        stack.pop()
    return sub


def run(file, out):
    try:
        html = ''
        try:
            with open(file) as fp:
                html += '<pre>' + ansi2html(fp.read()) + '</pre>'
        except IOError:
            print("File not found: %r" % 'pylint.txt', file=sys.stderr)

        with open(out, 'w') as out_fp:
            out_fp.write(html)
    except KeyboardInterrupt:
        pass



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import
import pickle
import itertools
import tarfile
import os
import sys
import numpy as np
from tqdm import tqdm
from nlp_architect.utils.generic import license_prompt
from nlp_architect.utils.io import download_unlicensed_file, valid_path_append


def pad_sentences(sentences, sentence_length=0, pad_val=0.):
    """
    Pad all sentences to have the same length (number of words)
    """
    lengths = [len(sent) for sent in sentences]

    nsamples = len(sentences)
    if sentence_length == 0:
        sentence_length = np.max(lengths)

    X = (np.ones((nsamples, sentence_length)) * pad_val).astype(dtype=np.int32)
    for i, sent in enumerate(sentences):
        trunc = sent[-sentence_length:]
        X[i, :len(trunc)] = trunc
    return X


def pad_stories(stories, sentence_length, max_story_length, pad_val=0.):
    """
    Pad all stories to have the same number of sentences (max_story_length).
    """
    nsamples = len(stories)

    X = (
        np.ones(
            (nsamples,
             max_story_length,
             sentence_length))
        * pad_val).astype(
            dtype=np.int32)

    for i, story in enumerate(stories):
        X[i, :len(story)] = story

    return X


class BABI_Dialog(object):
    """
    This class loads in the Facebook bAbI goal oriented dialog dataset and
    vectorizes them into user utterances, bot utterances, and answers.

    As described in: "Learning End-to-End Goal Oriented Dialog".
    https://arxiv.org/abs/1605.07683.

    For a particular task, the class will read both train and test files
    and combine the vocabulary.

    Args:
        path (str): Directory to store the dataset
        task (str): a particular task to solve (all bAbI tasks are train
            and tested separately)
        oov (bool, optional): Load test set with out of vocabulary entity words
        use_match_type (bool, optional): Flag to use match-type features
        use_time (bool, optional): Add time words to each memory, encoding when the
            memory was formed
        use_speaker_tag (bool, optional): Add speaker words to each memory
            (<BOT> or <USER>) indicating who spoke each memory.
        cache_match_type (bool, optional): Flag to save match-type features after processing
        cache_vectorized (bool, optional): Flag to save all vectorized data after processing

    Attributes:
        data_dict (dict): Dictionary containing final vectorized train, val, and test datasets
        cands (np.array): Vectorized array of potential candidate answers, encoded
        as integers, as returned by BABI_Dialog class. Shape = [num_cands, max_cand_length]
        num_cands (int): Number of potential candidate answers.
        max_cand_len (int): Maximum length of a candidate answer sentence in number of words.
        memory_size (int): Maximum number of sentences to keep in memory at any given time.
        max_utt_len (int): Maximum length of any given sentence / user utterance
        vocab_size (int): Number of unique words in the vocabulary + 2 (0 is reserved for
            a padding symbol, and 1 is reserved for OOV)
        use_match_type (bool, optional): Flag to use match-type features
        kb_ents_to_type (dict, optional): For use with match-type features, dictionary of
            entities found in the dataset mapping to their associated match-type
        kb_ents_to_cand_idxs (dict, optional): For use with match-type features, dictionary
            mapping from each entity in the  knowledge base to the set of indicies in the
            candidate_answers array that contain that entity.
        match_type_idxs (dict, optional): For use with match-type features, dictionary
            mapping from match-type to the associated fixed index of the candidate vector
            which indicated this match type.
    """
    def __init__(self, path='.', task=1, oov=False, use_match_type=False,
                 use_time=True, use_speaker_tag=True, cache_match_type=False,
                 cache_vectorized=False):
        self.url = 'http://www.thespermwhale.com/jaseweston/babi'
        self.size = 3032766
        self.filename = 'dialog-bAbI-tasks.tgz'
        self.path = path
        self.task = task - 1
        self.oov = oov
        self.use_match_type = use_match_type
        self.cache_vectorized = cache_vectorized
        self.match_type_vocab = None
        self.match_type_idxs = None

        self.tasks = [
            'dialog-babi-task1-API-calls-',
            'dialog-babi-task2-API-refine-',
            'dialog-babi-task3-options-',
            'dialog-babi-task4-phone-address-',
            'dialog-babi-task5-full-dialogs-',
            'dialog-babi-task6-dstc2-'
        ]

        print('Preparing bAbI-dialog dataset. Looking in ./%s' % path)

        assert task in range(
            1, 7), "given task is not in the bAbI-dialog dataset"

        print('Task is %s' % (self.tasks[self.task]))

        (self.train_file, self.dev_file, self.test_file, self.cand_file,
         self.kb_file, self.vocab_file, self.vectorized_file) = self.load_data()

        # Parse files into sets of dialogue and user/bot utterance pairs
        self.train_dialog = self.parse_dialog(
            self.train_file, use_time, use_speaker_tag)
        self.dev_dialog = self.parse_dialog(
            self.dev_file, use_time, use_speaker_tag)
        self.test_dialog = self.parse_dialog(
            self.test_file, use_time, use_speaker_tag)

        self.candidate_answers_w = self.load_candidate_answers()

        if self.use_match_type:
            self.kb_ents_to_type = self.load_kb()
            self.kb_ents_to_cand_idxs = self.create_match_maps()
        else:
            self.kb_ents_to_type = None
            self.kb_ents_to_cand_idxs = None

        self.compute_statistics()

        if self.use_match_type:
            self.encode_match_feats()

        if self.cache_vectorized and os.path.exists(self.vectorized_file):
            print('Loading cached vectorized data from: {}'.format(self.vectorized_file))
            with open(self.vectorized_file, 'rb') as f:
                (self.train, self.dev, self.test, self.cands) = pickle.load(f)

        else:
            self.train = self.vectorize_stories(self.train_dialog)
            self.dev = self.vectorize_stories(self.dev_dialog)
            self.test = self.vectorize_stories(self.test_dialog)
            self.cands = self.vectorize_cands(self.candidate_answers_w)

            if self.cache_vectorized:
                print('Caching vectorized data at {}'.format(self.vectorized_file))
                with open(self.vectorized_file, 'wb') as f:
                    pickle.dump((self.train, self.dev, self.test, self.cands), f)

        self.data_dict['train'] = {
            'memory': {
                'data': self.train[0], 'axes': ('batch', 'memory_axis', 'sentence_axis')},
            'memory_mask': {
                'data': self.train[1], 'axes': ('batch', 'memory_axis')},
            'user_utt': {
                'data': self.train[2], 'axes': ('batch', 'sentence_axis')},
            'answer': {
                'data': self.train[3], 'axes': ('batch', 'cand_axis')}}

        self.data_dict['dev'] = {
            'memory': {
                'data': self.dev[0], 'axes': ('batch', 'memory_axis', 'sentence_axis')},
            'memory_mask': {
                'data': self.dev[1], 'axes': ('batch', 'memory_axis')},
            'user_utt': {
                'data': self.dev[2], 'axes': ('batch', 'sentence_axis')},
            'answer': {
                'data': self.dev[3], 'axes': ('batch', 'cand_axis')}}

        self.data_dict['test'] = {
            'memory': {
                'data': self.test[0], 'axes': ('batch', 'memory_axis', 'sentence_axis')},
            'memory_mask': {
                'data': self.test[1], 'axes': ('batch', 'memory_axis')},
            'user_utt': {
                'data': self.test[2], 'axes': ('batch', 'sentence_axis')},
            'answer': {
                'data': self.test[3], 'axes': ('batch', 'cand_axis')}}

        # Add question-specific candidate answers if we are using match-type
        if self.use_match_type:
            self.data_dict['train']['cands_mat'] = {
                'data': self.create_cands_mat('train', cache_match_type),
                'axes': ('batch', 'cand_axis', 'REC')}
            self.data_dict['dev']['cands_mat'] = {
                'data': self.create_cands_mat('dev', cache_match_type),
                'axes': ('batch', 'cand_axis', 'REC')}
            self.data_dict['test']['cands_mat'] = {
                'data': self.create_cands_mat('test', cache_match_type),
                'axes': ('batch', 'cand_axis', 'REC')}

    def load_data(self):
        """
        Fetch and extract the Facebook bAbI-dialog dataset if not already downloaded.

        Returns:
            tuple: training and test filenames are returned
        """
        if self.task < 5:
            self.candidate_answer_filename = 'dialog-babi-candidates.txt'
            self.kb_filename = 'dialog-babi-kb-all.txt'
            self.cands_mat_filename = 'babi-cands-with-matchtype_{}.npy'
            self.vocab_filename = 'dialog-babi-vocab-task{}'.format(self.task + 1) +\
                                  '_matchtype{}.pkl'.format(self.use_match_type)
        else:
            self.candidate_answer_filename = 'dialog-babi-task6-dstc2-candidates.txt'
            self.kb_filename = 'dialog-babi-task6-dstc2-kb.txt'
            self.cands_mat_filename = 'dstc2-cands-with-matchtype_{}.npy'
            self.vocab_filename = 'dstc2-vocab-task{}_matchtype{}.pkl'.format(self.task + 1,
                                                                              self.use_match_type)

        self.vectorized_filename = 'vectorized_task{}.pkl'.format(self.task + 1)

        self.data_dict = {}
        self.vocab = None
        self.workdir, filepath = valid_path_append(
            self.path, '', self.filename)
        if not os.path.exists(filepath):
            if license_prompt('bAbI-dialog',
                              'https://research.fb.com/downloads/babi/',
                              self.path) is False:
                sys.exit(0)

            download_unlicensed_file(self.url, self.filename, filepath, self.size)

        self.babi_dir_name = self.filename.split('.')[0]

        self.candidate_answer_filename = self.babi_dir_name + \
            '/' + self.candidate_answer_filename
        self.kb_filename = self.babi_dir_name + '/' + self.kb_filename
        self.cands_mat_filename = os.path.join(
            self.workdir, self.babi_dir_name + '/' + self.cands_mat_filename)
        self.vocab_filename = self.babi_dir_name + '/' + self.vocab_filename
        self.vectorized_filename = self.babi_dir_name + '/' + self.vectorized_filename

        task_name = self.babi_dir_name + '/' + self.tasks[self.task] + '{}.txt'

        train_file = os.path.join(self.workdir, task_name.format('trn'))
        dev_file = os.path.join(self.workdir, task_name.format('dev'))
        test_file_postfix = 'tst-OOV' if self.oov else 'tst'
        test_file = os.path.join(
            self.workdir,
            task_name.format(test_file_postfix))

        cand_file = os.path.join(self.workdir, self.candidate_answer_filename)
        kb_file = os.path.join(self.workdir, self.kb_filename)
        vocab_file = os.path.join(self.workdir, self.vocab_filename)
        vectorized_file = os.path.join(self.workdir, self.vectorized_filename)

        if (os.path.exists(train_file) is False
                or os.path.exists(dev_file) is False
                or os.path.exists(test_file) is False
                or os.path.exists(cand_file) is False):
            with tarfile.open(filepath, 'r:gz') as f:
                f.extractall(self.workdir)

        return train_file, dev_file, test_file, cand_file, kb_file, vocab_file, vectorized_file

    @staticmethod
    def parse_dialog(fn, use_time=True, use_speaker_tag=True):
        """
        Given a dialog file, parse into user and bot utterances, adding time and speaker tags.

        Args:
            fn (str): Filename to parse
            use_time (bool, optional): Flag to append 'time-words' to the end of each utterance
            use_speaker_tag (bool, optional): Flag to append tags specifiying the speaker to
                each utterance.
        """
        with open(fn, 'r') as f:
            text = f.readlines()

        # Going to be filled with triplets of (memory, last user utterance,
        # desired bot utterance)
        all_dialogues = []
        current_memory = []

        for line in tqdm(text, desc="Parsing"):
            line = line.replace('\n', '')

            # End of dialgue
            if not line:
                current_memory = []
                continue

            number = line.split(' ')[0]

            if '\t' not in line:
                # Line is returned results form API call, store as memory in
                # current dialogue
                current_memory.append(line.split(' ') + ['<USER>'])
            else:
                user_utt, bot_utt = ' '.join(line.split(' ')[1:]).split('\t')

                # Split utterances into words so we can encode as BOW query
                user_utt_w = user_utt.split(' ')
                bot_utt_w = bot_utt.split(' ')

                # Add training example
                # Don't split bot utterance so we can directly compare with
                # candidate answers (and make onehot)
                all_dialogues.append(
                    (current_memory[:], user_utt_w[:], bot_utt))

                # Add time words and speaker tag
                if use_time:
                    user_utt_w += [str(number) + "_TIME"]
                    bot_utt_w += [str(number) + "_TIME"]
                if use_speaker_tag:
                    user_utt_w += ['<USER>']
                    bot_utt_w += ['<BOT>']

                # Add split and modified user and bot utterances to memory
                current_memory += [user_utt_w, bot_utt_w]

        return all_dialogues

    def words_to_vector(self, words):
        """
        Convert a list of words into vector form.

        Args:
            words (list) : List of words.

        Returns:
            list : Vectorized list of words.
        """
        return [self.word_to_index[w] if w in self.vocab else self.word_to_index[
            '<OOV>'] for w in words]

    def one_hot_vector(self, answer):
        """
        Create one-hot representation of an answer.

        Args:
            answer (string) : The word answer.

        Returns:
            list : One-hot representation of answer.
        """
        vector = np.zeros(self.num_cands)
        vector[self.candidate_answers.index(answer)] = 1
        return vector

    def vectorize_stories(self, data):
        """
        Convert (memory, user_utt, answer) word data into vectors.

        If sentence length < max_utt_len it is padded with 0's
        If memory length < memory size, it is padded with empty memorys (max_utt_len 0's)

        Args:
            data (tuple) : Tuple of memories, user_utt, answer word data.

        Returns:
            tuple : Tuple of memories, memory_lengths, user_utt, answer vectors.
        """
        m, ml, m_mask, u, a = [], [], [], [], []
        for mem, utt, answer in tqdm(data, desc="Vectorizing"):
            m.append([self.words_to_vector(sent) for sent in mem])
            ml.append(len(mem))
            mask_zero_len = self.memory_size - ml[-1]
            m_mask.append([1.0 for _ in range(ml[-1])]
                          + [0.0 for _ in range(mask_zero_len)])

            u.append(self.words_to_vector(utt))
            a.append(self.one_hot_vector(answer))

        m = np.array([pad_sentences(sents, self.max_utt_len) for sents in m])
        m = pad_stories(m, self.max_utt_len, self.memory_size)
        m_mask = np.array(m_mask)

        u = pad_sentences(u, self.max_utt_len)
        a = np.array(a)

        return (m, m_mask, u, a)

    def vectorize_cands(self, data):
        """
        Convert candidate answer word data into vectors.

        If sentence length < max_cand_len it is padded with 0's

        Args:
            data (list of lists) : list of candidate answers split into words

        Returns:
            tuple (2d numpy array): padded numpy array of word indexes forr all candidate answers
        """
        c = []
        for cand in data:
            c.append(self.words_to_vector(cand))

        c = pad_sentences(c, self.max_cand_len)
        return c

    def get_vocab(self, dialog):
        """
        Compute vocabulary from the set of dialogs.
        """
        # Extract only the memory words and user utterance words (these will
        # contain all vocab in the end)
        dialog_words = [x[0] + [x[1]] for x in dialog]

        # Concatenate separate dialogues
        all_utts = list(itertools.chain.from_iterable(dialog_words))
        # Concatenate all utterances to get list of words
        all_words = list(itertools.chain.from_iterable(all_utts))

        # Add candidate answer words
        all_words += list(itertools.chain.from_iterable(self.candidate_answers_w))

        if self.use_match_type:
            # Add match-type words
            self.match_type_vocab = list(set(self.kb_ents_to_type.values()))
            all_words += list(self.match_type_vocab) + \
                list(self.kb_ents_to_type.keys())
            # Also compute indicies into each candidate vector for the
            # different match types (fixed position)
            self.match_type_idxs = {
                mt: i + self.max_cand_len_pre_match for i,
                mt in enumerate(
                    self.match_type_vocab)}

        vocab = list(set(all_words))
        return vocab

    def compute_statistics(self):
        """
        Compute vocab, word index, and max length of stories and queries.
        """
        all_dialog = self.train_dialog + self.dev_dialog + self.test_dialog

        self.max_cand_len_pre_match = max(
            list(map(len, self.candidate_answers_w)))

        vocab = self.get_vocab(all_dialog)

        self.vocab = vocab
        # Reserve 0 for masking via pad_sequences, 1 for oov
        self.vocab_size = len(vocab) + 2

        if os.path.exists(self.vocab_file):
            with open(self.vocab_file, 'rb') as f:
                self.word_to_index = pickle.load(f)

            self.index_to_word = dict((self.word_to_index[k], k) for k in self.word_to_index)
        else:
            self.word_to_index = dict((c, i + 2) for i, c in enumerate(vocab))
            self.index_to_word = dict((i + 2, c) for i, c in enumerate(vocab))

            self.word_to_index['<PAD>'] = 0
            self.word_to_index['<OOV>'] = 1
            self.index_to_word[0] = ''  # empty so we dont print a bunch of padding
            self.index_to_word[1] = '<OOV>'

            with open(self.vocab_file, 'wb') as f:
                pickle.dump(self.word_to_index, f)

        memories = [m for m, _, _ in all_dialog]
        self.memory_size = max(list(map(len, memories)))

        dialog_words = [x[0] + [x[1]] for x in all_dialog]
        # Concatenate separate dialogues
        all_utts = list(itertools.chain.from_iterable(dialog_words))
        self.max_utt_len = max(list(map(len, all_utts)))

        self.num_cands = len(self.candidate_answers)

        if self.use_match_type:
            # Add num_match_types slots to each candidate answer so they can
            # be filled if we find a matching word
            self.max_cand_len = self.max_cand_len_pre_match + \
                len(self.match_type_vocab)
        else:
            self.max_cand_len = self.max_cand_len_pre_match

    @staticmethod
    def clean_cands(cand):
        """
        Remove leading line number and final newline from candidate answer
        """
        return ' '.join(cand.split(' ')[1:]).replace('\n', '')

    def load_candidate_answers(self):
        """
        Load candidate answers from file, compute number, and store for final softmax
        """
        with open(self.cand_file, 'r') as f:
            cands_text = f.readlines()
        self.candidate_answers = list(map(self.clean_cands, cands_text))

        # Create BOW representation of candidate answers for final prediction
        # softmax
        candidate_answers_w = list(
            map(lambda x: x.split(' '), self.candidate_answers))
        return candidate_answers_w

    def process_interactive(
            self,
            line_in,
            context,
            response,
            db_results,
            time_feat):
        """
        Parse a given user's input into the same format as training, build the memory
        from the given context and previous response, update the context.
        """
        # Parse user input
        line_in = line_in.replace('\n', '')
        line_w = line_in.split(' ')

        # Enocde user input and current context
        user_utt_w = self.words_to_vector(line_w)

        # Add last bot response to context before we create memory
        if response:
            bot_utt_w = response.split(' ')
            bot_utt_w += [str(time_feat - 1) + "_TIME", '<BOT>']
            context += [bot_utt_w]

        # If line_in is not silence, we are cutting off the api call
        # and making a correction
        if db_results and line_in == '<SILENCE>':
            for result in db_results:
                res_utt_w = [str(time_feat)] + result.split(' ') + ['<USER>']
                context += [res_utt_w]
                time_feat += 1

        # truncate context to max memory size
        context = context[-self.memory_size:]
        memory = [self.words_to_vector(sent) for sent in context]

        # Compute memory mask
        ml = len(memory)
        mask_zero_len = self.memory_size - ml
        m_mask = [1.0 for _ in range(ml)] + [0.0 for _ in range(mask_zero_len)]
        m_mask = np.array(m_mask)

        # Pad memory to max memory fize
        memory = pad_sentences(memory, self.max_utt_len)
        memory_pad = (
            np.zeros(
                (self.memory_size, self.max_utt_len))).astype(
                    dtype=np.int32)
        memory_pad[:len(memory)] = memory

        # Pad user utt to sentence size
        user_utt_pad = (np.zeros((self.max_utt_len,))).astype(dtype=np.int32)
        user_utt_trunc = user_utt_w[-self.max_utt_len:]
        user_utt_pad[:len(user_utt_trunc)] = user_utt_trunc

        # Add time features and store user utterance in context
        user_utt_w_mem = line_w + [str(time_feat) + "_TIME", '<USER>']
        context += [user_utt_w_mem]

        # Compute candidate_matrix if match features, otherwise just return
        # None & it'll use default
        if self.use_match_type:
            # dupliucate candidates for all examples
            cands_mat = np.array(self.cands)

            all_words = list(np.unique(memory.flatten())) + \
                list(np.unique(user_utt_trunc))

            # For each word, if it's an entity, add it's match-type word to the
            # candidate
            for word in all_words:
                if word in self.kb_ents_to_type.keys():
                    cand_idxs = self.kb_ents_to_cand_idxs[word]
                    ent_type_word = self.kb_ents_to_type[word]
                    match_type_word_idx = self.match_type_idxs[ent_type_word]

                    cands_mat[cand_idxs, match_type_word_idx] = ent_type_word
        else:
            cands_mat = np.array(self.cands)

        time_feat += 1

        return user_utt_pad, context, memory_pad, cands_mat, time_feat

    def load_kb(self):
        """
        Load knowledge base from file, parse into entities and types
        """
        with open(self.kb_file, 'r') as f:
            kb_text = f.readlines()

        if self.task < 5:
            # Split each kb entry into entity and entity type match, store as
            # dict
            kb_ents_to_type = {x.strip().split(
                '\t')[-1]: x.strip().split(' ')[2].split('\t')[-2] + "_MATCH" for x in kb_text}
        else:
            kb_ents_to_type = {
                x.strip().split(' ')[3]: x.strip().split(' ')[2] + "_MATCH" for x in kb_text}

        return kb_ents_to_type

    def create_match_maps(self):
        """
        Create dictionary mapping from each entity in the knowledge base to the set of
        indicies in the candidate_answers array that contain that entity. Will be used for
        quickly adding the match type features to the candidate answers during fprop.
        """

        kb_ents = self.kb_ents_to_type.keys()

        kb_ents_to_cand_idxs = {}

        for ent in kb_ents:
            kb_ents_to_cand_idxs[ent] = []

            for c_idx, cand in enumerate(self.candidate_answers_w):
                if ent in cand:
                    kb_ents_to_cand_idxs[ent].append(c_idx)

        return kb_ents_to_cand_idxs

    def encode_match_feats(self):
        """
        Replace entity names and match type names with indexes
        """
        self.kb_ents_to_type = {
            self.word_to_index[k]: self.word_to_index[v] for k,
            v in self.kb_ents_to_type.items()}
        self.kb_ents_to_cand_idxs = {
            self.word_to_index[k]: v for k,
            v in self.kb_ents_to_cand_idxs.items()}
        self.match_type_idxs = {
            self.word_to_index[k]: v for k,
            v in self.match_type_idxs.items()}

    def create_cands_mat(self, data_split, cache_match_type):
        """
        Add match type features to candidate answers for each example in the dataaset.
        Caches once complete.
        """
        cands_mat_filename = self.cands_mat_filename.format(data_split)

        data_dict = self.data_dict[data_split]

        # Returned cached matric if it exists
        if os.path.exists(cands_mat_filename):
            return np.load(cands_mat_filename)

        ndata = data_dict['user_utt']['data'].shape[0]
        # dupliucate candidates for all examples
        cands_mat = np.array([self.cands] * ndata)

        print("Adding match type features to {} set".format(data_split))

        for idx, (mem, utt) in enumerate(
                tqdm(zip(data_dict['memory']['data'], data_dict['user_utt']['data']))):
            # Get list of unique words currently in memory of user utt
            all_words = list(np.unique(mem.flatten())) + list(np.unique(utt))

            # For each word, if it's an entity, add it's match-type word to the
            # candidate
            for word in all_words:
                if word in self.kb_ents_to_type.keys():
                    cand_idxs = self.kb_ents_to_cand_idxs[word]
                    ent_type_word = self.kb_ents_to_type[word]
                    match_type_word_idx = self.match_type_idxs[ent_type_word]

                    cands_mat[idx][
                        cand_idxs, match_type_word_idx] = ent_type_word

        if cache_match_type:
            # Cache computed matrix
            print(
                "Saving candidate matrix for {} to {}".format(
                    data_split,
                    cands_mat_filename))
            np.save(cands_mat_filename, cands_mat)

        return cands_mat



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division
from __future__ import print_function
import numpy as np
import tensorflow as tf
import re
from collections import Counter

from nlp_architect.utils.text import SpacyInstance


class MatchLSTMAnswerPointer(object):
    """
    Defines end to end MatchLSTM and Answer_Pointer network for Reading Comprehension
    """

    def __init__(self, params_dict, embeddings):
        """
        Args:
            params_dict: Dictionary containing the following keys-
                         'max_question' : max length of all questions in the dataset
                         'max_para' :  max length of all paragraphs in the dataset
                         'hidden_size': number of hidden units in the network
                         'batch_size' : batch size defined by user

            embeddings: Glove pretrained embedding matrix
        """

        # Assign Variables:
        self.max_question = params_dict['max_question']
        self.max_para = params_dict['max_para']
        self.hidden_size = params_dict['hidden_size']
        self.batch_size = params_dict['batch_size']
        self.embeddings = embeddings
        self.inference_only = params_dict['inference_only']
        self.G_i = None
        self.attn = None
        self.stacked_lists_forward = None
        self.stacked_lists_reverse = None
        self.logits_withsf = None

        # init tokenizer
        self.tokenizer = SpacyInstance(disable=['tagger', 'ner', 'parser', 'vectors', 'textcat'])

        # Create Placeholders
        # Question ids
        self.question_ids = tf.placeholder(tf.int32, shape=[None, self.max_question],
                                           name="question_ids")
        # Paragraph ids
        self.para_ids = tf.placeholder(tf.int32, shape=[None, self.max_para],
                                       name="para_ids")
        # Length of question
        self.question_length = tf.placeholder(tf.int32, shape=[None],
                                              name="question_len")
        # Length of paragraph
        self.para_length = tf.placeholder(tf.int32, shape=[None],
                                          name="para_len")
        # Mask for paragraph
        self.para_mask = tf.placeholder(tf.float32, shape=[None, self.max_para],
                                        name="para_mask")
        # Mask for question
        self.ques_mask = tf.placeholder(tf.float32, shape=[None, self.max_question],
                                        name="ques_mask")
        # Answer spans
        if self.inference_only is False:
            self.labels = tf.placeholder(tf.int32, shape=[None, 2], name="labels")
        # Dropout value
        self.dropout = tf.placeholder(tf.float32, shape=[], name="dropout")
        self.global_step = tf.Variable(0, name='global')

        # Get variables
        self.create_variables()

        # Define model
        self.create_model()

    def create_variables(self):
        """
        Function to create variables used for training

        """
        # define all variables required for training
        self.W_p = tf.get_variable("W_p", [1, self.hidden_size, self.hidden_size])
        self.W_r = tf.get_variable("W_r", [1, self.hidden_size, self.hidden_size])
        self.W_q = tf.get_variable("W_q", [1, self.hidden_size, self.hidden_size])
        self.w_lr = tf.get_variable("w_lr", [1, 1, self.hidden_size])
        self.b_p = tf.get_variable("b_p", [1, self.hidden_size, 1])
        self.c_p = tf.get_variable("c_p", [1])
        self.ones_vector = tf.constant(np.ones([1, self.max_question]), dtype=tf.float32)

        self.ones_vector_exp = tf.tile(tf.expand_dims(self.ones_vector, 0),
                                       [self.batch_size, 1, 1])

        self.ones_vector_para = tf.constant(np.ones([1, self.max_para]), dtype=tf.float32)

        self.ones_para_exp = tf.tile(tf.expand_dims(self.ones_vector_para, 0),
                                     [self.batch_size, 1, 1])

        self.ones_embed = tf.tile(tf.expand_dims(tf.constant(
                                  np.ones([1, self.hidden_size]), dtype=tf.float32), 0),
                                  [self.batch_size, 1, 1])

        self.V_r = tf.get_variable("V_r", [1, self.hidden_size, 2 * self.hidden_size])
        self.W_a = tf.get_variable("W_a", [1, self.hidden_size, self.hidden_size])
        self.b_a = tf.get_variable("b_a", [1, self.hidden_size, 1])
        self.v_a_pointer = tf.get_variable("v_a_pointer", [1, 1, self.hidden_size])
        self.c_pointer = tf.get_variable("c_pointer", [1, 1, 1])
        self.Wans_q = tf.get_variable("Wans_q", [1, self.hidden_size, self.hidden_size])
        self.Wans_v = tf.get_variable("Wans_v", [1, self.hidden_size, self.hidden_size])
        self.Vans_r = tf.get_variable("Vans_r", [1, self.hidden_size, self.max_question])

        self.mask_ques_mul = tf.matmul(tf.transpose(self.ones_embed, [0, 2, 1]),
                                       tf.expand_dims(self.ques_mask, 1))

    def create_model(self):
        """
        Function to set up the end 2 end reading comprehension model

        """
        # Embedding Layer
        embedding_lookup = tf.Variable(self.embeddings, name="word_embeddings",
                                       dtype=tf.float32, trainable=False)
        # Embedding Lookups
        self.question_emb = tf.nn.embedding_lookup(embedding_lookup, self.question_ids,
                                                   name="question_embed")

        self.para_emb = tf.nn.embedding_lookup(embedding_lookup, self.para_ids,
                                               name="para_embed")

        # Apply dropout after embeddings
        self.question = tf.nn.dropout(self.question_emb, self.dropout)
        self.para = tf.nn.dropout(self.para_emb, self.dropout)

        # Encoding Layer
        # Share weights of pre-processing LSTM layer with both para and
        # question
        with tf.variable_scope("encoded_question"):
            self.lstm_cell_question = tf.nn.rnn_cell.BasicLSTMCell(self.hidden_size,
                                                                   state_is_tuple=True)

            self.encoded_question, _ = tf.nn.dynamic_rnn(self.lstm_cell_question,
                                                         self.question,
                                                         self.question_length,
                                                         dtype=tf.float32)

        with tf.variable_scope("encoded_para"):
            self.encoded_para, _ = tf.nn.dynamic_rnn(self.lstm_cell_question, self.para,
                                                     self.para_length, dtype=tf.float32)

        # Define Match LSTM and Answer Pointer Cells
        with tf.variable_scope("match_lstm_cell"):
            self.match_lstm_cell = tf.nn.rnn_cell.BasicLSTMCell(self.hidden_size,
                                                                state_is_tuple=True)

        with tf.variable_scope("answer_pointer_cell"):
            self.lstm_pointer_cell = tf.nn.rnn_cell.BasicLSTMCell(self.hidden_size,
                                                                  state_is_tuple=True)

        print('Match LSTM Pass')
        # Match LSTM Pass in forward direction
        self.unroll_with_attention(reverse=False)
        self.encoded_para_reverse = tf.reverse(self.encoded_para, axis=[1])
        # Match LSTM Pass in reverse direction
        self.unroll_with_attention(reverse=True)
        # Apply dropout
        self.stacked_lists = tf.concat([tf.nn.dropout(self.stacked_lists_forward,
                                                      tf.maximum(self.dropout, 0.8)),
                                        tf.nn.dropout(self.stacked_lists_reverse,
                                                      tf.maximum(self.dropout, 0.8))], 1)

        # Answer pointer pass
        self.logits = self.answer_pointer_pass()
        if self.inference_only is False:
            print('Settting up Loss')
            # Compute Losses
            loss_1 = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=self.logits[0],
                                                                    labels=self.labels[:, 0])

            loss_2 = tf.nn.sparse_softmax_cross_entropy_with_logits(logits=self.logits[1],
                                                                    labels=self.labels[:, 1])
            # Total Loss
            self.loss = tf.reduce_mean(loss_1 + loss_2)
            self.learning_rate = tf.constant(0.002)

            print('Set up optmizer')
            # Optmizer
            self.optimizer = tf.train.AdamOptimizer(self.learning_rate). \
                minimize(self.loss, global_step=self.global_step)

    def unroll_with_attention(self, reverse=False):
        """
        Function to run the match_lstm pass in both forward and reverse directions

        Args:
        reverse: Boolean indicating whether to unroll in reverse directions

        """
        # Intitialze first hidden_state with zeros
        h_r_old = tf.constant(
            np.zeros([self.batch_size, self.hidden_size, 1]), dtype=tf.float32)
        final_state_list = []

        for i in range(self.max_para):

            if not reverse:
                encoded_paraslice = tf.gather(self.encoded_para, indices=i, axis=1)
            else:
                encoded_paraslice = tf.gather(self.encoded_para_reverse, indices=i,
                                              axis=1)

            W_p_expanded = tf.tile(self.W_p, [self.batch_size, 1, 1])
            W_q_expanded = tf.tile(self.W_q, [self.batch_size, 1, 1])
            W_r_expanded = tf.tile(self.W_r, [self.batch_size, 1, 1])
            w_lr_expanded = tf.tile(self.w_lr, [self.batch_size, 1, 1])
            b_p_expanded = tf.tile(self.b_p, [self.batch_size, 1, 1])

            int_sum = tf.matmul(W_p_expanded, tf.expand_dims(encoded_paraslice, 2)) + \
                tf.matmul(W_r_expanded, h_r_old) + b_p_expanded

            int_sum_new = tf.matmul(int_sum, tf.expand_dims(self.ques_mask, 1))

            int_sum1 = tf.matmul(W_q_expanded,
                                 tf.transpose(self.encoded_question, [0, 2, 1]))

            self.G_i = tf.nn.tanh(int_sum_new + int_sum1) + \
                tf.expand_dims(self.c_p * self.ques_mask, 1)

            # Attention Vector
            self.attn = tf.nn.softmax(tf.matmul(w_lr_expanded, self.G_i))

            z1 = encoded_paraslice

            z2 = tf.squeeze(tf.matmul(tf.transpose(self.encoded_question, [0, 2, 1]),
                                      tf.transpose(self.attn, [0, 2, 1])), axis=2)

            z_i_stacked = tf.concat([z1, z2], 1)
            if i == 0:
                h_r_old, cell_state_old = self.match_lstm_cell(
                    z_i_stacked, state=self.match_lstm_cell.zero_state(self.batch_size,
                                                                       dtype=tf.float32))
            else:
                h_r_old, cell_state_old = self.match_lstm_cell(z_i_stacked,
                                                               state=cell_state_old)

            final_state_list.append(h_r_old)
            h_r_old = tf.expand_dims(h_r_old, 2)
            stacked_lists = tf.stack(final_state_list, 1)

        if not reverse:
            # Mask Output
            mask_mult_lstm_forward = tf.matmul(tf.transpose(self.ones_embed, [0, 2, 1]),
                                               tf.expand_dims(self.para_mask, 1))

            self.stacked_lists_forward = tf.multiply(tf.transpose(stacked_lists, [0, 2, 1]),
                                                     mask_mult_lstm_forward)
        else:
            # Mask Output
            mask_mult_lstm_reverse = tf.matmul(tf.transpose(
                self.ones_embed, [0, 2, 1]), tf.expand_dims(tf.reverse(
                                                            self.para_mask, axis=[1]), 1))

            self.stacked_lists_reverse = tf.reverse(tf.multiply(tf.transpose(
                stacked_lists, [0, 2, 1]), mask_mult_lstm_reverse), axis=[2])

    def answer_pointer_pass(self):
        """
        Function to run the answer pointer pass:

        Args:
            None

        Returns:
            List of logits for start and end indices of the answer
        """

        V_r_expanded = tf.tile(self.V_r, [self.batch_size, 1, 1])
        W_a_expanded = tf.tile(self.W_a, [self.batch_size, 1, 1])
        b_a_expanded = tf.tile(self.b_a, [self.batch_size, 1, 1])
        mask_multiplier_1 = tf.expand_dims(self.para_mask, 1)
        mask_multiplier = self.ones_para_exp

        v_apointer_exp = tf.tile(self.v_a_pointer, [self.batch_size, 1, 1])

        # Zero initialization
        h_k_old = tf.constant(
            np.zeros([self.batch_size, self.hidden_size, 1]), dtype=tf.float32)

        b_k_lists = []

        print("Answer Pointer Pass")

        for i in range(0, 2):
            sum1 = tf.matmul(V_r_expanded, self.stacked_lists)
            sum2 = tf.matmul(W_a_expanded, h_k_old) + b_a_expanded
            F_k = tf.nn.tanh(sum1 + tf.matmul(sum2, mask_multiplier))

            b_k_withoutsf = tf.matmul(v_apointer_exp, F_k)

            b_k = tf.nn.softmax(b_k_withoutsf + tf.log(mask_multiplier_1))
            lstm_cell_inp = tf.squeeze(tf.matmul(self.stacked_lists,
                                       tf.transpose(b_k, [0, 2, 1])), axis=2)

            with tf.variable_scope("lstm_pointer"):
                if i == 0:
                    h_k_old, cell_state_pointer = self.lstm_pointer_cell(
                        lstm_cell_inp, state=self.lstm_pointer_cell.zero_state(self.batch_size,
                                                                               dtype=tf.float32))
                else:
                    h_k_old, cell_state_pointer = self.lstm_pointer_cell(lstm_cell_inp,
                                                                         state=cell_state_pointer)

            h_k_old = tf.expand_dims(h_k_old, 2)
            b_k_lists.append(b_k_withoutsf + tf.log(mask_multiplier_1))

        self.logits_withsf = [tf.nn.softmax(tf.squeeze(b_k_lists[0], axis=1)),
                              tf.nn.softmax(tf.squeeze(b_k_lists[1], axis=1))]

        return [tf.squeeze(b_k_lists[0], axis=1), tf.squeeze(b_k_lists[1], axis=1)]

    @staticmethod
    def obtain_indices(preds_start, preds_end):
        """
        Function to get answer indices given the predictions

        Args:
            preds_start: predicted start indices
            predictions: predicted end indices

        Returns:
            final start and end indices for the answer
        """
        ans_start = []
        ans_end = []
        for i in range(preds_start.shape[0]):
            max_ans_id = -100000000
            st_idx = 0
            en_idx = 0
            ele1 = preds_start[i]
            ele2 = preds_end[i]
            len_para = len(ele1)
            for j in range(len_para):
                for k in range(15):
                    if j + k >= len_para:
                        break
                    ans_start_int = ele1[j]
                    ans_end_int = ele2[j + k]
                    if (ans_start_int + ans_end_int) > max_ans_id:
                        max_ans_id = ans_start_int + ans_end_int
                        st_idx = j
                        en_idx = j + k

            ans_start.append(st_idx)
            ans_end.append(en_idx)

        return (np.array(ans_start), np.array(ans_end))

    def cal_f1_score(self, ground_truths, predictions):
        """
        Function to calculate F-1 and EM scores

        Args:
            ground_truths: labels given in the dataset
            predictions: logits predicted by the network

        Returns:
            F1 score and Exact-Match score
        """
        start_idx, end_idx = self.obtain_indices(predictions[0], predictions[1])
        f1 = 0
        exact_match = 0
        for i in range(self.batch_size):
            ele1 = start_idx[i]
            ele2 = end_idx[i]
            preds = np.linspace(ele1, ele2, abs(ele2 - ele1 + 1))
            length_gts = abs(ground_truths[i][1] - ground_truths[i][0] + 1)
            gts = np.linspace(ground_truths[i][0], ground_truths[i][1], length_gts)
            common = Counter(preds) & Counter(gts)
            num_same = sum(common.values())

            exact_match += int(np.array_equal(preds, gts))
            if num_same == 0:
                f1 += 0
            else:
                precision = 1.0 * num_same / len(preds)
                recall = 1.0 * num_same / len(gts)
                f1 += (2 * precision * recall) / (precision + recall)

        return 100 * (f1 / self.batch_size), 100 * (exact_match / self.batch_size)

    def get_dynamic_feed_params(self, question_str, vocab_reverse):
        """
        Function to get required feed_dict format for user entered questions.
        Used mainly in the demo mode.

        Args:
           question_str: question string
           vocab_reverse: vocab dictionary with words as keys and indices as values

        Returns:
           question_idx: list of indicies represnting the question padded to max length
           question_len: actual length of the question
           ques_mask: mask for question_idx

        """

        question_words = [word.replace("``", '"').replace("''", '"')
                          for word in self.tokenizer.tokenize(question_str)]

        question_ids = [vocab_reverse[ele] for ele in question_words]

        if len(question_ids) < self.max_question:
            pad_length = self.max_question - len(question_ids)
            question_idx = question_ids + [0] * pad_length
            question_len = len(question_ids)
            ques_mask = np.zeros([1, self.max_question])
            ques_mask[0, 0:question_len] = 1
            ques_mask = ques_mask.tolist()[0]

        return question_idx, question_len, ques_mask

    def run_loop(self, session, train, mode='train', dropout=0.6):
        """
        Function to run training/validation loop and display training loss, F1 & EM scores

        Args:
            session: tensorflow session
            train:   data dictionary for training/validation
            dropout: float value
            mode: 'train'/'val'
        """

        nbatches = int((len(train['para']) / self.batch_size))
        f1_score = 0
        em_score = 0
        for idx in range(nbatches):
            # Train for all batches
            start_batch = self.batch_size * idx
            end_batch = self.batch_size * (idx + 1)
            if end_batch > len(train['para']):
                break

            # Create feed dictionary
            feed_dict_qa = {
                self.para_ids: np.asarray(train['para'][start_batch:end_batch]),
                self.question_ids: np.asarray(train['question'][start_batch:end_batch]),
                self.para_length: np.asarray(train['para_len'][start_batch:end_batch]),
                self.question_length: np.asarray(train['question_len'][start_batch:end_batch]),
                self.labels: np.asarray(train['answer'][start_batch:end_batch]),
                self.para_mask: np.asarray(train['para_mask'][start_batch:end_batch]),
                self.ques_mask: np.asarray(train['question_mask'][start_batch:end_batch]),
                self.dropout: dropout
            }
            # Training Phase
            if mode == 'train':
                _, train_loss, _, logits, labels = session.run(
                    [self.optimizer, self.loss, self.learning_rate, self.logits_withsf,
                     self.labels], feed_dict=feed_dict_qa)

                if (idx % 20 == 0):
                    print('iteration = {}, train loss = {}'.format(idx, train_loss))
                    f1_score, em_score = self.cal_f1_score(labels, logits)
                    print("F-1 and EM Scores are", f1_score, em_score)

                self.global_step.assign(self.global_step + 1)

            else:
                logits, labels = session.run([self.logits_withsf, self.labels],
                                             feed_dict=feed_dict_qa)

                f1_score_int, em_score_int = self.cal_f1_score(labels, logits)
                f1_score += f1_score_int
                em_score += em_score_int

        # Validation Phase
        if mode == 'val':
            print(
                "Validation F1 and EM scores are",
                f1_score / nbatches,
                em_score / nbatches)

    # pylint: disable=inconsistent-return-statements
    def inference_mode(self, session, valid, vocab_tuple, num_examples, dropout=1.0,
                       dynamic_question_mode=False, dynamic_usr_question="",
                       dynamic_question_index=0):
        """
          Function to run inference_mode for reading comprehension

          Args:
              session: tensorflow session
              valid: data dictionary for validation set
              vocab_tuple: a tuple containing voacab dictionaries in forward and reverse directions
              num_examples : specify the number of samples to run for inference
              dropout : Float value which is always 1.0 for inference
              dynamic_question_mode : boolean to enable whether or not accept
                                      questions from the user(used in the demo mode)

        """
        vocab_forward = vocab_tuple[0]
        vocab_reverse = vocab_tuple[1]

        for idx in range(num_examples):
            if dynamic_question_mode is True:
                idx = dynamic_question_index
                required_params = self.get_dynamic_feed_params(dynamic_usr_question, vocab_reverse)
                question_ids = required_params[0]
                question_length = required_params[1]
                ques_mask = required_params[2]
                test_paragraph = [vocab_forward[ele] for ele in valid[idx][0] if ele != 0]
                para_string = " ".join(map(str, test_paragraph))
            else:
                # Print Paragraph
                print("\n")
                print("Paragraph Number:", idx)
                test_paragraph = [vocab_forward[ele] for ele in valid[idx][0] if ele != 0]
                para_string = " ".join(map(str, test_paragraph))
                print(re.sub(r'\s([?.!,"](?:\s|$))', r'\1', para_string))

                # Print corresponding Question
                test_question = [vocab_forward[ele] for ele in valid[idx][1] if ele != 0]
                ques_string = " ".join(map(str, test_question))
                print("Question:", re.sub(r'\s([?.!"",])', r'\1', ques_string))
                question_ids = valid[idx][1]
                question_length = valid[idx][3]
                ques_mask = valid[idx][6]

            # Create a feed dictionary
            feed_dict_qa = {self.para_ids: np.expand_dims(valid[idx][0], 0),
                            self.question_ids: np.expand_dims(question_ids, 0),
                            self.para_length: np.expand_dims(valid[idx][2], 0),
                            self.question_length: np.expand_dims(question_length, 0),
                            self.para_mask: np.expand_dims(valid[idx][5], 0),
                            self.ques_mask: np.expand_dims(ques_mask, 0),
                            self.dropout: dropout}

            # Run session and obtain indices
            predictions = session.run([self.logits_withsf], feed_dict=feed_dict_qa)

            # Get the start and end indices of the answer
            start_idx, end_idx = self.obtain_indices(predictions[0][0], predictions[0][1])
            answer_ind = valid[idx][0][start_idx[0]:end_idx[0] + 1]

            # Print answer
            req_ans = [vocab_forward[ele] for ele in answer_ind if ele != 0]
            ans_string = " ".join(map(str, req_ans))
            answer = re.sub(r'\s([?.!",])', r'\1', ans_string)
            print("Answer:", answer)
            if dynamic_question_mode is True:
                return {"answer": answer}



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import os
from typing import Dict, Set

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType, \
    OnlineOROfflineMethod
from nlp_architect.utils.io import load_json_file
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


class VerboceanRelationExtraction(RelationExtraction):
    def __init__(self, method: OnlineOROfflineMethod = OnlineOROfflineMethod.ONLINE,
                 vo_file: str = None):
        """
        Extract Relation between two mentions according to VerbOcean knowledge

        Args:
            method (optional): OnlineOROfflineMethod.{ONLINE/OFFLINE} run against full VerbOcean or
                a sub-set of it (default = ONLINE)
            vo_file (required): str Location of VerbOcean file to work with
        """
        logger.info('Loading Verb Ocean module')
        if vo_file is not None and os.path.isfile(vo_file):
            if method == OnlineOROfflineMethod.OFFLINE:
                self.vo = load_json_file(vo_file)
            elif method == OnlineOROfflineMethod.ONLINE:
                self.vo = self.load_verbocean_file(vo_file)
            logger.info('Verb Ocean module lead successfully')
        else:
            raise FileNotFoundError('VerbOcean file not found or not in path..')
        super(VerboceanRelationExtraction, self).__init__()

    def extract_all_relations(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> Set[RelationType]:
        ret_ = set()
        ret_.add(self.extract_sub_relations(mention_x, mention_y, RelationType.VERBOCEAN_MATCH))
        return ret_

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        if relation is not RelationType.VERBOCEAN_MATCH:
            return RelationType.NO_RELATION_FOUND

        mention_x_str = mention_x.tokens_str
        mention_y_str = mention_y.tokens_str
        if StringUtils.is_pronoun(mention_x_str.lower()) or StringUtils.is_pronoun(
                mention_y_str.lower()):
            return RelationType.NO_RELATION_FOUND

        if self.is_verbocean_relation(mention_x, mention_y):
            return RelationType.VERBOCEAN_MATCH

        return RelationType.NO_RELATION_FOUND

    def is_verbocean_relation(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> bool:
        """
        Check if input mentions has VerbOcean relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            bool
        """
        x_head = mention_x.mention_head

        y_head = mention_y.mention_head

        rel = None

        if x_head in self.vo and y_head in self.vo[x_head]:
            rel = self.vo[x_head][y_head]
        elif y_head in self.vo and x_head in self.vo[y_head]:
            rel = self.vo[y_head][x_head]

        match_result = False
        if rel is not None and rel != "[unk]" and rel != "[low-vol]":
            match_result = True

        return match_result

    @staticmethod
    def get_supported_relations():
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.VERBOCEAN_MATCH]

    @staticmethod
    def load_verbocean_file(fname: str) -> Dict[str, Dict[str, str]]:
        """
        Method to load referent dictionary to memory

        Returns:
            List[RelationType]
        """
        word_dict = {}
        with open(fname) as f:
            for line in f:
                word1, rel, word2, _, _ = line.strip().split()
                if word1 not in word_dict:
                    word_dict[word1] = {}
                word_dict[word1][word2] = rel
        return word_dict



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# pylint: disable=c-extension-no-member
""" REST Server to respond to different API requests """
import gzip
import json
from os import path

import hug
from falcon import status_codes

from nlp_architect.server.service import Service, parse_headers, format_response

services = {}

api = hug.API(__name__)
api.http.add_middleware(hug.middleware.CORSMiddleware(api, max_age=10))


def prefetch_models():
    models = ['machine_comprehension', 'bist', 'ner', 'intent_extraction']
    for model in models:
        services[model] = Service(model)


@hug.get('/comprehension_paragraphs')
def get_paragraphs():
    if not services['machine_comprehension']:
        services['machine_comprehension'] = Service('machine_comprehension')
    return services['machine_comprehension'].get_paragraphs()


# pylint: disable=inconsistent-return-statements
@hug.post()
def inference(request, body, response):
    """Makes an inference to a certain model"""
    print(body)
    if request.headers.get('CONTENT-TYPE') == 'application/gzip':
        try:
            original_data = gzip.decompress(request.stream.read())
            input_docs = json.loads(str(original_data, 'utf-8'))["docs"]
            model_name = json.loads(str(original_data, 'utf-8'))["model_name"]
        except Exception:
            response.status = hug.HTTP_500
            return {'status': 'unexpected gzip error'}
    elif request.headers.get('CONTENT-TYPE') == 'application/json':
        if isinstance(body, str):
            body = json.loads(body)
        model_name = body.get('model_name')
        input_docs = body.get('docs')
    else:
        response.status = status_codes.HTTP_400
        return {'status': 'Content-Type header must be application/json or application/gzip'}
    if not model_name:
        response.status = status_codes.HTTP_400
        return {'status': 'model_name is required'}
    # If we've already initialized it, no use in reinitializing
    if not services.get(model_name):
        services[model_name] = Service(model_name)
    if not isinstance(input_docs, list):  # check if it's an array instead
        response.status = status_codes.HTTP_400
        return {'status': 'request not in proper format '}
    headers = parse_headers(request.headers)
    parsed_doc = services[model_name].get_service_inference(input_docs, headers)
    resp_format = request.headers["RESPONSE-FORMAT"]
    ret = format_response(resp_format, parsed_doc)
    if request.headers.get('CONTENT-TYPE') == 'application/gzip':
        response.content_type = resp_format
        response.body = ret
        # no return due to the fact that hug seems to assume json type upon return
    else:
        return ret


@hug.static('/')
def static():
    """Statically serves a directory to client"""
    return [path.join(path.dirname(path.realpath(__file__)), 'angular-ui/dist/angular-ui')]
    # return [os.path.realpath(os.path.join('./', 'server/angular-ui/dist/angular-ui'))]


@hug.get(['/home', '/visual/{page}', '/annotate/{page}', '/machine_comprehension'],
         output=hug.output_format.file)
def get_index():
    index = path.join(path.dirname(path.realpath(__file__)),
                      'angular-ui/dist/angular-ui/index.html')
    return index


prefetch_models()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
""" Service file used to import different features for server """
import json
import logging
import os.path
from importlib import import_module

from nlp_architect.utils.io import gzip_str

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def format_response(resp_format, parsed_doc):
    """
    Transform string of server's response to the requested format

    Args:
        resp_format(str): the desired response format
        parsed_doc: the server's response

    Returns:
        formatted response
    """
    logger.info('preparing response JSON')
    ret = None
    if (resp_format == "json") or ('json' in resp_format) or (not resp_format):
        # if not specified resp_format then default is json
        ret = parsed_doc
    if resp_format == "gzip" or 'gzip' in resp_format:
        ret = gzip_str(parsed_doc)
    return ret


def parse_headers(req_headers):
    """
    Load headers from request to dictionary

    Args:
        req_headers (dict): the request's headers

    Returns:
        dict: dictionary hosting the request headers
    """
    headers_lst = ["CONTENT-TYPE", "CONTENT-ENCODING", "RESPONSE-FORMAT",
                   "CLEAN", "DISPLAY-POST-PREPROCCES",
                   "DISPLAY-TOKENS", "DISPLAY-TOKEN-TEXT", "IS-HTML"]
    headers = {}
    for header_tag in headers_lst:
        if header_tag in req_headers:
            headers[header_tag] = req_headers[header_tag]
        else:
            headers[header_tag] = None
    return headers


def set_headers(res):
    """
    set expected headers for request (CORS)

    Args:
        res (:obj:`falcon.Response`): the request
    """
    res.set_header('Access-Control-Allow-Origin', '*')
    res.set_header("Access-Control-Allow-Credentials", "true")
    res.set_header('Access-Control-Allow-Methods', "GET,HEAD,OPTIONS,POST,PUT")
    res.set_header('Access-Control-Allow-Headers',
                   "Access-Control-Allow-Headers, Access-Control-Allow-Origin,"
                   " Origin,Accept, X-Requested-With, Content-Type, "
                   "Access-Control-Request-Method, "
                   "Access-Control-Request-Headers, Response-Format, clean, "
                   "display-post-preprocces, display-tokens, "
                   "display-token-text")


def package_home(gdict):
    """
    help function for running paths from out-of-class scripts
    """
    filename = gdict["__file__"]
    return os.path.dirname(filename)


def extract_module_name(model_path):
    """
    Extract the module's name from path

    Args:
        model_path(str): the module's class path

    Returns:
        str: the modules name
    """
    class_name = "".join(model_path.split(".")[0].title().split("_"))
    return class_name


class Service(object):
    """Handles loading and inference using specific models"""
    def __init__(self, service_name):
        self.service_type = None
        self.is_spacy = False
        self.service = self.load_service(service_name)

    def get_paragraphs(self):
        return self.service.get_paragraphs()

    # pylint: disable=eval-used
    def get_service_inference(self, docs, headers):
        """
        get parser response from service API

        Args:
            headers (list(str)): the headers of the request
            docs: input received from the request

        Returns:
            the service API output
        """
        logger.info('sending documents to parser')
        response_data = []
        for i, doc in enumerate(docs):
            inference_doc = self.service.inference(doc["doc"])
            if self.is_spacy is True:
                parsed_doc = inference_doc.displacy_doc()
                doc_dic = {"id": doc["id"], "doc": parsed_doc}
                # Potentially a security risk
                if headers['IS-HTML'] is not None and eval(headers['IS-HTML']):
                    # a visualizer requestadd type of service (core/annotate) to response
                    doc_dic["type"] = self.service_type
                response_data.append(doc_dic)
            else:
                inference_doc['id'] = i + 1
                response_data.append(inference_doc)
        return response_data

    def load_service(self, name):
        """
        Initialize and load service from input given name, using "services.json" properties file

        Args:
            name (str):
                The name of service to upload using server

        Returns:
            The loaded service
        """
        with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "services.json")) \
                as prop_file:
            properties = json.load(prop_file)
        folder_path = properties["api_folders_path"]
        service_name_error = "'{0}' is not an existing service - " \
                             "please try using another service.".format(name)
        if name in properties:
            model_relative_path = properties[name]["file_name"]
        else:
            logger.error(service_name_error)
            raise Exception(service_name_error)
        if not model_relative_path:
            logger.error(service_name_error)
            raise Exception(service_name_error)
        module_path = ".".join(model_relative_path.split(".")[:-1])
        module_name = extract_module_name(model_relative_path)
        module = import_module(folder_path + module_path)
        class_api = getattr(module, module_name)
        upload_service = class_api()
        upload_service.load_model()
        self.service_type = properties[name]["type"]
        self.is_spacy = properties[name].get('spacy', False)
        return upload_service



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import logging
import sys
from os import path, makedirs

from tqdm import tqdm

from nlp_architect.pipelines.spacy_np_annotator import NPAnnotator, get_noun_phrases
from nlp_architect.solutions.trend_analysis.scoring_utils import TextSpanScoring
from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.io import download_unlicensed_file
from nlp_architect.utils.text import SpacyInstance

nlp_chunker_url = 'https://s3-us-west-2.amazonaws.com/nlp-architect-data/models/chunker/'
chunker_model_dat_file = 'model_info.dat.params'
chunker_model_file = 'model.h5'
chunker_local_path = str(LIBRARY_OUT / 'chunker-pretrained')
logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


class NPScorer(object):
    def __init__(self, parser=None):
        if parser is None:
            self.nlp = SpacyInstance(
                disable=['ner', 'parser', 'vectors', 'textcat']).parser
        else:
            self.nlp = parser

        self.nlp.add_pipe(self.nlp.create_pipe('sentencizer'), first=True)
        _path_to_model = path.join(chunker_local_path, chunker_model_file)
        if not path.exists(chunker_local_path):
            makedirs(chunker_local_path)
        if not path.exists(_path_to_model):
            logger.info(
                'The pre-trained model to be downloaded for NLP Architect word'
                ' chunker model is licensed under Apache 2.0')
            download_unlicensed_file(nlp_chunker_url, chunker_model_file, _path_to_model)
        _path_to_params = path.join(chunker_local_path, chunker_model_dat_file)
        if not path.exists(_path_to_params):
            download_unlicensed_file(nlp_chunker_url, chunker_model_dat_file, _path_to_params)
        self.nlp.add_pipe(NPAnnotator.load(_path_to_model, _path_to_params), last=True)

    def score_documents(self, texts: list, limit=-1, return_all=False, min_tf=5):
        documents = []
        assert len(texts) > 0, 'texts should contain at least 1 document'
        assert min_tf > 0, 'min_tf should be at least 1'
        with tqdm(total=len(texts), desc='documents scoring progress', unit='docs') as pbar:
            for doc in self.nlp.pipe(texts, n_threads=-1):
                if len(doc) > 0:
                    documents.append(doc)
                pbar.update(1)

        corpus = []
        for doc in documents:
            spans = get_noun_phrases(doc)
            if len(spans) > 0:
                corpus.append((doc, spans))

        if len(corpus) < 1:
            return []

        documents, doc_phrases = list(zip(*corpus))
        scorer = TextSpanScoring(documents=documents, spans=doc_phrases, min_tf=min_tf)
        tfidf_scored_list = scorer.get_tfidf_scores()
        if len(tfidf_scored_list) < 1:
            return []
        cvalue_scored_list = scorer.get_cvalue_scores()
        freq_scored_list = scorer.get_freq_scores()

        if limit > 0:
            tf = {tuple(k[0]): k[1] for k in tfidf_scored_list}
            cv = {tuple(k[0]): k[1] for k in cvalue_scored_list}
            fr = {tuple(k[0]): k[1] for k in freq_scored_list}
            tfidf_scored_list_limit = []
            cvalue_scored_list_limit = []
            freq_scored_list_limit = []
            for phrase in list(zip(*tfidf_scored_list))[0][:limit]:
                tfidf_scored_list_limit.append((phrase, tf[tuple(phrase)]))
                cvalue_scored_list_limit.append((phrase, cv[tuple(phrase)]))
                freq_scored_list_limit.append((phrase, fr[tuple(phrase)]))
            tfidf_scored_list = tfidf_scored_list_limit
            cvalue_scored_list = cvalue_scored_list_limit
            freq_scored_list = freq_scored_list_limit

        tfidf_scored_list = scorer.normalize_l2(tfidf_scored_list)
        cvalue_scored_list = scorer.normalize_l2(cvalue_scored_list)
        freq_scored_list = scorer.normalize_minmax(freq_scored_list, invert=True)
        tfidf_scored_list = scorer.normalize_minmax(tfidf_scored_list)
        cvalue_scored_list = scorer.normalize_minmax(cvalue_scored_list)
        if return_all:
            tf = {tuple(k[0]): k[1] for k in tfidf_scored_list}
            cv = {tuple(k[0]): k[1] for k in cvalue_scored_list}
            fr = {tuple(k[0]): k[1] for k in freq_scored_list}
            final_list = []
            for phrases in tf.keys():
                final_list.append(([p for p in phrases], tf[phrases], cv[phrases], fr[phrases]))
            return final_list
        merged_list = scorer.interpolate_scores([tfidf_scored_list, cvalue_scored_list],
                                                [0.5, 0.5])
        merged_list = scorer.multiply_scores([merged_list, freq_scored_list])
        merged_list = scorer.normalize_minmax(merged_list)
        final_list = []
        for phrases, score in merged_list:
            if any([len(p) > 1 for p in phrases]):
                final_list.append(([p for p in phrases], score))
        return final_list
